"""GFF format detection and auto read/write dispatch (binary, JSON, XML)."""

from __future__ import annotations

import os
import struct

from pathlib import Path
from typing import TYPE_CHECKING

from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.gff.gff_data import GFFContent
from pykotor.resource.formats.gff.io_gff import GFFBinaryReader, GFFBinaryWriter
from pykotor.resource.formats.gff.io_gff_json import GFFJSONReader, GFFJSONWriter
from pykotor.resource.formats.gff.io_gff_xml import GFFXMLReader, GFFXMLWriter
from pykotor.resource.type import ResourceType

if TYPE_CHECKING:
    from pykotor.resource.formats.gff.gff_data import GFF
    from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES

_UTF8_BOM: bytes = b"\xef\xbb\xbf"


def _adjust_offset_after_utf8_bom(
    source: SOURCE_TYPES,
    offset: int,
    size: int | None,
) -> tuple[int, int | None]:
    """Skip a leading UTF-8 BOM at (source, offset) so binary GFF readers align on the file type id."""
    if isinstance(source, (bytes, bytearray, memoryview)):
        total: int = len(source)
        end: int = total if size is None else min(total, offset + size)
        if offset + 3 > end:
            return offset, size
        head = source[offset : offset + 3]
        if isinstance(head, memoryview):
            head = head.tobytes()
        elif isinstance(head, bytearray):
            head = bytes(head)
        if head != _UTF8_BOM:
            return offset, size
        new_offset = offset + 3
        if size is None:
            return new_offset, None
        return new_offset, size - 3

    if isinstance(source, (str, os.PathLike)):
        try:
            with Path(os.fspath(source)).open("rb") as handle:
                handle.seek(offset)
                bom_candidate = handle.read(3)
        except OSError:
            return offset, size
        if bom_candidate != _UTF8_BOM:
            return offset, size
        new_offset = offset + 3
        if size is None:
            return new_offset, None
        if size < 3:
            return offset, size
        return new_offset, size - 3

    return offset, size


def _is_binary_gff_v32_magic(magic8: bytes) -> bool:
    """True if bytes are a KotOR-style binary GFF magic (4-char type + b'V3.2')."""
    if len(magic8) < 8:
        return False
    type_id = magic8[:4].decode("windows-1252", errors="ignore")
    if not any(x.value == type_id for x in GFFContent):
        return False
    return magic8[4:8] == b"V3.2"


def _is_binary_gff_v32_at(peek: bytes, index: int) -> bool:
    """True if a KotOR-style binary GFF header (4-char type + b'V3.2') starts at index."""
    if index < 0 or index + 8 > len(peek):
        return False
    return _is_binary_gff_v32_magic(peek[index : index + 8])


def _gff_payload_span(source: SOURCE_TYPES, offset: int, size: int | None) -> int:
    """Length in bytes available from (source, offset) when constrained by size."""
    if isinstance(source, (bytes, bytearray, memoryview)):
        total: int = len(source)
        end: int = total if size is None else min(total, offset + size)
        return max(0, end - offset)
    if isinstance(source, (str, os.PathLike)):
        try:
            st = Path(os.fspath(source)).stat()
        except OSError:
            return 0
        if size is None:
            return max(0, st.st_size - offset)
        return max(0, min(st.st_size, offset + size) - offset)
    return 0


def _binary_gff_header_offsets_plausible(header48: bytes, span_from_header_start: int) -> bool:
    """True if the 48 bytes after GFF magic describe table offsets that fit within span_from_header_start."""
    if span_from_header_start < 56 or len(header48) < 48:
        return False
    (
        struct_off,
        struct_cnt,
        field_off,
        field_cnt,
        label_off,
        label_cnt,
        field_data_off,
        _field_data_cnt,
        field_idx_off,
        field_idx_cnt,
        list_idx_off,
        list_idx_cnt,
    ) = struct.unpack_from("<12I", header48, 0)
    ptrs = (
        struct_off,
        field_off,
        label_off,
        field_data_off,
        field_idx_off,
        list_idx_off,
    )
    for p in ptrs:
        if p < 56 or p >= span_from_header_start:
            return False
    if max(struct_cnt, field_cnt, label_cnt, field_idx_cnt, list_idx_cnt) > 0x100000:
        return False
    if label_off + label_cnt * 16 > span_from_header_start:
        return False
    if struct_off + struct_cnt * 12 > span_from_header_start:
        return False
    if field_off + field_cnt * 12 > span_from_header_start:
        return False
    # field_indices_count / list_indices_count often overstate reserved space; the reader does not
    # load those arrays wholesale. Bounds above match io_gff.load()'s upfront reads.
    return True


def _locate_binary_gff_header_structural(
    source: SOURCE_TYPES,
    offset: int,
    size: int | None,
    *,
    max_scan: int = 256,
) -> int | None:
    """Find binary GFF V3.2 start by validating header tables against stream span (allows non-padding prefixes)."""
    span = _gff_payload_span(source, offset, size)
    if span < 56:
        return None
    window_len = min(span, max_scan + 56)
    buf: bytes
    if isinstance(source, (bytes, bytearray, memoryview)):
        raw = memoryview(source)[offset : offset + window_len]
        buf = raw.tobytes()
    elif isinstance(source, (str, os.PathLike)):
        try:
            with Path(os.fspath(source)).open("rb") as handle:
                handle.seek(offset)
                buf = handle.read(window_len)
        except OSError:
            return None
    else:
        return None
    if len(buf) < 56:
        return None
    upper = min(max_scan, max(0, span - 56))
    for i in range(upper + 1):
        if i + 56 > len(buf):
            break
        if not _is_binary_gff_v32_magic(buf[i : i + 8]):
            continue
        header48 = buf[i + 8 : i + 56]
        if _binary_gff_header_offsets_plausible(header48, span - i):
            return i
    return None


def _leading_bytes_allow_gff_resync(peek: bytes, header_off: int) -> bool:
    """True if bytes before a candidate GFF header are only harmless padding (BOM / WS / NUL).

    Without this, a coincidental b\"ARE \"+b\"V3.2\" deep in arbitrary binary can be mistaken for
    a real header and produce corrupt parses (OSError: exceed stream boundaries).
    """
    if header_off <= 0:
        return True
    if header_off > len(peek):
        return False
    prefix = peek[:header_off]
    idx = 0
    if len(prefix) >= 3 and prefix[:3] == _UTF8_BOM:
        idx = 3
    allowed = frozenset(b" \t\r\n\x00")
    return all(b in allowed for b in prefix[idx:])


def _locate_binary_gff_header(peek: bytes, *, max_scan: int = 64) -> int | None:
    """Return byte offset of binary GFF V3.2 header within peek, or None if not found."""
    upper = min(max_scan + 1, max(0, len(peek) - 7))
    for i in range(upper):
        if not _leading_bytes_allow_gff_resync(peek, i):
            continue
        if _is_binary_gff_v32_at(peek, i):
            return i
    return None


def _classify_gff_peek(peek: bytes) -> ResourceType:
    """Infer GFF container format from the first bytes (handles UTF-8 BOM and leading whitespace)."""
    if len(peek) < 4:
        return ResourceType.INVALID
    if peek.startswith(_UTF8_BOM):
        peek = peek[len(_UTF8_BOM) :]
    if len(peek) < 4:
        return ResourceType.INVALID
    first4 = peek[:4].decode("windows-1252", errors="ignore")
    if len(peek) >= 8 and any(x.value == first4 for x in GFFContent) and peek[4:8] == b"V3.2":
        return ResourceType.GFF
    stripped = peek.lstrip(b" \t\r\n")
    if not stripped:
        return ResourceType.INVALID
    if stripped[:1] == b"<":
        return ResourceType.GFF_XML
    if stripped[:1] == b"{":
        return ResourceType.GFF_JSON
    return ResourceType.INVALID


def detect_gff(
    source: SOURCE_TYPES,
    offset: int = 0,
) -> ResourceType:
    """Returns what format the GFF data is believed to be in.

    This function performs a basic check and does not guarantee accuracy of the result or integrity of the data.

    Args:
    ----
        source: Source of the GFF data.
        offset: Offset into the data.

    Raises:
    ------
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.

    Returns:
    -------
        The format of the GFF data.
    """
    file_format: ResourceType
    try:
        with BinaryReader.from_auto(source, offset) as reader:
            remain = reader.remaining()
            chunk_len = min(512, max(0, remain))
            peek = reader.read_bytes(chunk_len)
        file_format = _classify_gff_peek(peek)
        if file_format == ResourceType.INVALID and _locate_binary_gff_header(peek) is not None:
            file_format = ResourceType.GFF
        elif file_format == ResourceType.INVALID and _locate_binary_gff_header_structural(source, offset, None) is not None:
            file_format = ResourceType.GFF
    except (FileNotFoundError, PermissionError, IsADirectoryError):
        raise
    except OSError:
        file_format = ResourceType.INVALID

    return file_format


def read_gff(
    source: SOURCE_TYPES,
    offset: int = 0,
    size: int | None = None,
    file_format: ResourceType | None = None,
) -> GFF:  # sourcery skip: hoist-statement-from-if, reintroduce-else
    """Returns an GFF instance from the source.

    The file format (GFF or GFF_XML) is automatically determined before parsing the data.

    Args:
    ----
        source: The source of the data.
        offset: The byte offset of the file inside the data.
        size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.
        file_format: The file format to use (ResourceType.GFF or ResourceType.GFF_XML or ResourceType.GFF_JSON). If not specified, it will be detected automatically.

    Raises:
    ------
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.
        ValueError: If the file was corrupted or the format could not be determined.

    Returns:
    -------
        A GFF instance.
    """
    offset, size = _adjust_offset_after_utf8_bom(source, offset, size)

    peek: bytes
    try:
        with BinaryReader.from_auto(source, offset) as reader:
            chunk_len = min(512, max(0, reader.remaining()))
            peek = reader.read_bytes(chunk_len)
    except (FileNotFoundError, PermissionError, IsADirectoryError):
        raise
    except OSError:
        peek = b""

    header_off = _locate_binary_gff_header(peek)
    if header_off is None:
        header_off = _locate_binary_gff_header_structural(source, offset, size)
    promoted_binary: bool = False
    if file_format is None:
        file_format = _classify_gff_peek(peek)
        if file_format == ResourceType.INVALID and header_off is not None:
            file_format = ResourceType.GFF
            promoted_binary = True
    if file_format == ResourceType.GFF and header_off is not None and (promoted_binary or header_off > 0):
        offset += header_off
        if size is not None:
            size = max(0, size - header_off)

    if file_format == ResourceType.GFF:
        return GFFBinaryReader(source, offset, size or 0).load()
    if file_format == ResourceType.GFF_XML:
        return GFFXMLReader(source, offset, size or 0).load()
    if file_format == ResourceType.GFF_JSON:
        return GFFJSONReader(source, offset, size or 0).load()

    msg = "Failed to determine the format of the GFF file."
    # if file_format == ResourceType.INVALID:
    raise ValueError(msg)


def write_gff(
    gff: GFF,
    target: TARGET_TYPES,
    file_format: ResourceType = ResourceType.GFF,
):
    """Writes the GFF data to the target location with the specified format (GFF or GFF_XML).

    Args:
    ----
        gff: The GFF file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
    ------
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be written to the specified destination.
        ValueError: If the specified format was unsupported.
    """
    if file_format.is_gff():
        GFFBinaryWriter(gff, target).write()
    elif file_format.name.endswith("_XML") and file_format.target_type().is_gff():
        GFFXMLWriter(gff, target).write()
    elif file_format.name.endswith("_JSON") and file_format.target_type().is_gff():
        GFFJSONWriter(gff, target).write()
    else:
        msg = "Unsupported format specified; use GFF, GFF_XML, or GFF_JSON."
        raise ValueError(msg)


def bytes_gff(
    gff: GFF,
    file_format: ResourceType = ResourceType.GFF,
) -> bytes:
    """Returns the GFF data in the specified format (GFF or GFF_XML) as a bytes object.

    This is a convenience method that wraps the write_gff() and read_gff() methods.

    Args:
    ----
        gff: GFF: The target GFF.
        file_format: The file format.

    Raises:
    ------
        ValueError: If the specified format was unsupported.

    Returns:
    -------
        The GFF data.
    """
    data = bytearray()
    write_gff(gff, data, file_format)
    return bytes(data)



"""Chitin (chitin.key + BIF) loader: read-only access to game resource index and data."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pykotor.common.stream import BinaryReader
from pykotor.extract.file import FileResource, ResourceIdentifier
from pykotor.resource.type import ResourceType
from pykotor.tools.path import CaseAwarePath

if TYPE_CHECKING:
    import os

    from collections.abc import Iterator
    from pathlib import Path

    from pykotor.common.misc import Game


class Chitin:
    """Chitin object is used for loading the list of resources stored in the chitin.key/.bif files used by the game.

    Chitin support is read-only and you cannot write your own key/bif files with this class yet.

    Reverse Engineering Notes:
    -------------------------
    The KOTOR engine uses CExoResMan (Resource Manager) to handle multiple archive types:
    - FIXED (0x00000000): KEY/BIF files (chitin.key + data/*.bif)
    - DIRECTORY (0x80000000): Loose files in directories
    - ERF (0x40000000): ERF/RIM archives (modules/*.rim, modules/*.erf)
    - RIM (0x20000000): RIM archives (specifically for texture packs)

    CExoResMan::AddKeyTable manages archive loading with these type flags.
    CExoResMan::ReadResource handles the actual resource loading from archives.

    References:
    ----------
        Based on swkotor.exe ERF structure:
        - See pykotor.resource.formats.erf.erf_data for addresses (K1 + TSL TODO). CExoEncapsulatedFile, AddEncapsulatedContents.
        Original BioWare engine binaries


    """

    KEY_ELEMENT_SIZE = 8

    def __init__(
        self,
        key_path: os.PathLike | str,
        base_path: os.PathLike | str | None = None,
        game: Game | None = None,
    ):
        self._key_path: CaseAwarePath = CaseAwarePath(key_path)
        base_path = self._key_path.parent if base_path is None else base_path
        self._base_path: CaseAwarePath = CaseAwarePath(base_path)

        self._resources: list[FileResource] = []
        self._resource_dict: dict[str, list[FileResource]] = {}
        self.game: Game | None = game
        self.load()

    def __iter__(self) -> Iterator[FileResource]:
        yield from self._resources

    def __len__(self):
        return len(self._resources)

    def load(self):
        """Reload the list of resource info linked from the chitin.key file."""
        self._resources.clear()
        self._resource_dict.clear()

        keys: dict[int, str]
        bifs: list[str]
        keys, bifs = self._get_chitin_data()
        for bif in bifs:  # Loop through all bifs in the chitin.key
            self._resource_dict[bif] = []
            absolute_bif_path: Path = self._base_path.joinpath(bif)
            if self.game is not None and self.game.is_ios():  # For some reason, the chitin.key references the .bif path instead of the correct .bzf path.
                absolute_bif_path = absolute_bif_path.with_suffix(".bzf")
            self.read_bif(absolute_bif_path, keys, bif)

    def read_bif(
        self,
        bif_path: Path,
        keys: dict[int, str],
        bif_filename: str,
    ):
        """Reads a single .bif file and adds its resources to this Chitin object.

        Args:
        ----
            bif_path: The path to the .bif file.
            keys: The dictionary mapping resource IDs to resrefs.
            bif_filename: The filename of the .bif file (used as a key in the resource dictionary).

        Returns:
        -------
            None
        """

        with BinaryReader.from_file(bif_path) as reader:
            _bif_file_type: str = reader.read_string(4)  # 0x0
            _bif_file_version: str = reader.read_string(4)  # 0x4
            resource_count: int = reader.read_uint32()  # 0x8
            _fixed_resource_count: int = reader.read_uint32()  # unimplemented/padding (always 0x00000000?)
            resource_offset: int = reader.read_uint32()  # 0x10 always the value hex 0x14 (dec 20)
            reader.seek(resource_offset)  # Skip to 0x14

            for _ in range(resource_count):
                # Initialize the FileResource and add to this chitin object's collections.
                resource = FileResource(
                    resname=keys[reader.read_uint32()],  # resref str
                    offset=reader.read_uint32(),
                    size=reader.read_uint32(),
                    restype=ResourceType.from_id(reader.read_uint32()),
                    filepath=bif_path,
                )
                self._resources.append(resource)
                self._resource_dict[bif_filename].append(resource)

    def _get_chitin_data(self) -> tuple[dict[int, str], list[str]]:
        with BinaryReader.from_file(self._key_path) as reader:
            # _key_file_type = reader.read_string(4)  # noqa: ERA001
            # _key_file_version = reader.read_string(4)  # noqa: ERA001
            reader.skip(8)
            bif_count: int = reader.read_uint32()
            key_count: int = reader.read_uint32()
            file_table_offset: int = reader.read_uint32()
            reader.skip(4)  # key table offset uint32

            files: list[tuple[int, int]] = []
            reader.seek(file_table_offset)
            for _ in range(bif_count):
                reader.skip(4)  # ??? 0x000696E0 in k1, 0x000DDD8A in k2
                file_offset: int = reader.read_uint32()
                file_length: int = reader.read_uint16()
                reader.skip(2)  # ??? 0x0001 in K1, 0x0000 in K2
                files.append((file_offset, file_length))

            bifs: list[str] = []
            for file_offset, file_length in files:
                reader.seek(file_offset)
                bif: str = reader.read_string(file_length)
                bifs.append(bif)

            keys: dict[int, str] = {}
            for _ in range(key_count):
                resref: str = reader.read_string(16)
                reader.skip(2)  # restype_id uint16
                res_id: int = reader.read_uint32()
                keys[res_id] = resref

            return keys, bifs

    def resource(
        self,
        resref: str,
        restype: ResourceType,
    ) -> bytes | None:
        """Returns the bytes data of the specified resource. If the resource does not exist then returns None instead.

        Args:
        ----
            resref: The resource ResRef.
            restype: The resource type.

        Returns:
        -------
            None or bytes data of resource.
        """
        query = ResourceIdentifier(resref, restype)
        resource: FileResource | None = next(
            (resource for resource in self._resources if resource == query),
            None,
        )
        return None if resource is None else resource.data()



"""This module handles classes relating to editing LTR files.

LTR (Letter) files contain Markov chain probability data for generating random names
during character creation. They use a 3rd-order Markov chain model with single-letter,
double-letter (bigram), and triple-letter (trigram) probability tables. Each table
stores probability values for characters appearing at the start, middle, or end of names.

References:
----------
    Based on swkotor.exe LTR structure:
    - LoadLTR - Loads LTR file for name generation
      * Parses binary LTR format with "LTR V1.0" header
      * Reads letter count (uint8, typically 28)
      * Reads single-letter probability arrays (84 bytes: 28 chars * 3 positions * 4 bytes)
      * Reads double-letter probability arrays (2352 bytes: 28 * 28 * 3 * 4)
      * Reads triple-letter probability arrays (65856 bytes: 28 * 28 * 28 * 3 * 4)
      * Builds Markov chain probability tables for random name generation
    - "LTR " file type identifier - First 4 bytes of LTR files
    - "V1.0" version identifier - Bytes 4-7 of LTR files
    - Letter Count field at offset 0x08 (1 byte, uint8) - Number of characters in set (26 or 28)
    - Single Letters start at offset 0x09 (84 bytes: 28 floats * 3 positions)
    - Double Letters start at offset 0x5D (2352 bytes: 28 * 28 * 3 * 4)
    - Triple Letters start at offset 0x965 (65856 bytes: 28 * 28 * 28 * 3 * 4)
    - ".ltr" extension - LTR file extension
    - Original BioWare engine binaries (TODO: Verify with REVA when available)
    https://github.com/mtijanic/nwn-misc/blob/master/nwnltr.c - Original C reference implementation
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:19-210
        Binary Format:
        -------------
        Header (9 bytes):
        Offset | Size | Type   | Description
        -------|------|--------|-------------
        0x00   | 4    | char[] | File Type ("LTR ")
        0x04   | 4    | char[] | File Version ("V1.0")
        0x08   | 1    | uint8  | Letter Count (26 or 28)
        Single Letters (84 bytes = 28 * 3 * 4):
        Start probabilities: 28 floats (4 bytes each)
        Middle probabilities: 28 floats (4 bytes each)
        End probabilities: 28 floats (4 bytes each)
        Double Letters (2352 bytes = 28 * 28 * 3 * 4):
        28 LetterSets, each containing:
        Start probabilities: 28 floats
        Middle probabilities: 28 floats
        End probabilities: 28 floats
        Triple Letters (65856 bytes = 28 * 28 * 28 * 3 * 4):
        28x28 LetterSets, each containing:
        Start probabilities: 28 floats
        Middle probabilities: 28 floats
        End probabilities: 28 floats
"""

from __future__ import annotations

import random
import secrets
import string

from pykotor.resource.formats._base import ComparableMixin
from pykotor.resource.type import ResourceType


class LTR(ComparableMixin):
    """Represents a LTR (Letter) file containing Markov chain name generation data.

    LTR files use 3rd-order Markov chains to generate random names. The probability
    tables store likelihood values for characters appearing in different positions
    (start, middle, end) based on previous character context (none, one, or two chars).

    References:
    ----------
    Based on swkotor.exe LTR structure:
    - LoadLTR - Loads LTR file for name generation
      * Parses binary LTR format with "LTR V1.0" header
      * Reads letter count (uint8, typically 28 for KotOR)
      * Reads single-letter probability arrays (84 bytes: 28 chars * 3 positions * 4 bytes per float)
      * Reads double-letter probability arrays (2352 bytes: 28 * 28 * 3 * 4)
      * Reads triple-letter probability arrays (65856 bytes: 28 * 28 * 28 * 3 * 4)
      * Builds 3rd-order Markov chain probability tables for random name generation
    - "LTR " file type identifier - First 4 bytes of LTR files (offset 0x00)
    - "V1.0" version identifier - Bytes 4-7 of LTR files (offset 0x04)
    - Letter Count field at offset 0x08 (1 byte, uint8) - Number of characters in set (26 for NWN, 28 for KotOR)
    - Single Letters start at offset 0x09 (84 bytes: 28 floats * 3 positions)
      * Start probabilities: 28 floats (offset 0x09-0x70)
      * Middle probabilities: 28 floats (offset 0x71-0xD8)
      * End probabilities: 28 floats (offset 0xD9-0x140)
    - Double Letters start at offset 0x5D (2352 bytes: 28 * 28 * 3 * 4)
      * 28 LetterSets, each with start/middle/end probabilities (28 floats each)
    - Triple Letters start at offset 0x965 (65856 bytes: 28 * 28 * 28 * 3 * 4)
      * 28x28 LetterSets, each with start/middle/end probabilities (28 floats each)
    - ".ltr" extension - LTR file extension identifier
    - Original BioWare engine binaries (TODO: Verify with REVA when available)
    Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:19-210



    Attributes:
    ----------
        CHARACTER_SET: String of valid characters (28 chars: a-z + apostrophe + hyphen)
            Reference: https://github.com/th3w1zard1/KotOR.js/tree/master/LTRObject.ts:23 (CharacterArrays[28])
            KotOR uses 28-character set: "abcdefghijklmnopqrstuvwxyz'-"
            NWN uses 26-character set: "abcdefghijklmnopqrstuvwxyz"

        NUM_CHARACTERS: Number of characters in character set (28 for KotOR)
            Fixed at 28 for KotOR games

        _singles: Single-letter probability block (no context)
            Reference: https://github.com/th3w1zard1/KotOR.js/tree/master/LTRObject.ts:31 (singleArray[3][28])
            Contains start/middle/end probabilities for each character
            Used to generate the first character of names

        _doubles: Double-letter probability blocks (1-character context)
            Reference: https://github.com/th3w1zard1/KotOR.js/tree/master/LTRObject.ts:32 (doubleArray[28][3][28])
            Array of 28 LetterSets, indexed by previous character
            Used to generate second character based on first character

        _triples: Triple-letter probability blocks (2-character context)
            Reference: https://github.com/th3w1zard1/KotOR.js/tree/master/LTRObject.ts:33 (tripleArray[28][28][3][28])
            28x28 array of LetterSets, indexed by previous two characters
            Used to generate third and subsequent characters based on previous two
    """

    CHARACTER_SET = string.ascii_lowercase + "'-"
    # O(1) lookup for character index; avoids repeated list.index() in hot paths.
    _CHAR_INDEX: dict[str, int] = {c: i for i, c in enumerate(string.ascii_lowercase + "'-")}
    NUM_CHARACTERS = 28

    BINARY_TYPE = ResourceType.LTR
    COMPARABLE_FIELDS = ("_singles", "_doubles", "_triples")

    def __init__(self):
        # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:31

        # Single-letter probability block (no context, for first character)
        self._singles: LTRBlock = LTRBlock(LTR.NUM_CHARACTERS)

        # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:32

        # Double-letter probability blocks (1-character context, for second character)
        # Array of 28 blocks, indexed by previous character
        self._doubles: list[LTRBlock] = [LTRBlock(LTR.NUM_CHARACTERS) for _ in range(LTR.NUM_CHARACTERS)]

        # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:33

        # Triple-letter probability blocks (2-character context, for third+ characters)
        # 28x28 array of blocks, indexed by previous two characters
        self._triples: list[list[LTRBlock]] = [[LTRBlock(LTR.NUM_CHARACTERS) for _ in range(LTR.NUM_CHARACTERS)] for _ in range(LTR.NUM_CHARACTERS)]

    def __eq__(self, other):
        if not isinstance(other, LTR):
            return NotImplemented  # type: ignore[no-any-return]
        return self._singles == other._singles and self._doubles == other._doubles and self._triples == other._triples

    def __hash__(self):
        return hash((self._singles, tuple(self._doubles), tuple(tuple(row) for row in self._triples)))

    @staticmethod
    def _chance() -> float:
        """Returns a randomly generated float between 0.0 and 1.0 inclusive.

        Returns:
        -------
            A float between 0.0 and 1.0 inclusive.
        """
        return secrets.randbelow(1001) / 1000

    def generate(
        self,
        seed: int | None = None,
    ) -> str:
        """Returns a randomly generated name based on the LTR instance data.

        This method implements a 3rd-order Markov chain name generation algorithm.
        It generates names character-by-character using probability tables, starting
        with single-letter probabilities, then double-letter (bigram), then triple-letter
        (trigram) probabilities for subsequent characters.

        References:
        ----------
        See module docstring for engine addresses (K1 + TSL TODO).
        https://github.com/mtijanic/nwn-misc/blob/master/nwnltr.c - Original C reference
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:128-210 (getName method)



        Args:
        ----
            seed: Randomness seed for reproducible name generation.

        Returns:
        -------
            A randomly generated name (capitalized).

        Algorithm:
        ---------
            1. Generate first character using single-letter start probabilities
            2. Generate second character using double-letter start probabilities (based on first char)
            3. Generate third character using triple-letter start probabilities (based on first two chars)
            4. Generate subsequent characters using triple-letter middle probabilities
            5. Terminate when triple-letter end probability is selected or max attempts reached
        """
        # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:134 (prob = Math.random()) (prob = Math.random()
        # Set random seed for reproducible generation
        random.seed(seed)

        done = False

        while not done:
            attempts = 0
            name: str = ""

            # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:145-152

            # Generate first character using single-letter start probabilities
            for char in LTR.CHARACTER_SET:
                if LTR._chance() < self._singles.get_start(char):
                    name += char
                    break
            else:
                continue

            # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:154-161

            # Generate second character using double-letter start probabilities (indexed by first char)
            for char in LTR.CHARACTER_SET:
                index = LTR._CHAR_INDEX[name[-1]]
                if LTR._chance() < self._doubles[index].get_start(char):
                    name += char
                    break
            else:
                continue

            # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:163-170

            # Generate third character using triple-letter start probabilities (indexed by first two chars)
            for char in LTR.CHARACTER_SET:
                index1 = LTR._CHAR_INDEX[name[-2]]
                index2 = LTR._CHAR_INDEX[name[-1]]
                if LTR._chance() < self._triples[index1][index2].get_start(char):
                    name += char
                    break
            else:
                continue

            # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:173-200

            # Generate subsequent characters using triple-letter middle/end probabilities
            while True:
                prob: float = LTR._chance()

                # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:175 (Math.floor(Math.random() * 2147483647) % 12) (Math.floor(Math.random()

                # Check if name should end (probability increases with name length)
                if (secrets.randbelow(12) % 12) <= len(name):
                    # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:176-182

                    # Select final character using triple-letter end probabilities
                    for char in LTR.CHARACTER_SET:
                        index1 = LTR._CHAR_INDEX[name[-2]]
                        index2 = LTR._CHAR_INDEX[name[-1]]
                        if prob < self._triples[index1][index2].get_end(char):
                            name += char
                            return name.capitalize()

                # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:190-195

                # Generate next character using triple-letter middle probabilities
                for char in LTR.CHARACTER_SET:
                    index1 = LTR._CHAR_INDEX[name[-2]]
                    index2 = LTR._CHAR_INDEX[name[-1]]
                    if prob < self._triples[index1][index2].get_middle(char):
                        name += char
                        break
                else:
                    # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:197-200

                    # No valid character found - increment attempts and check termination
                    attempts += 1
                    if len(name) < 4 or attempts > 100:  # noqa: PLR2004
                        break

        msg = f"Unknown problem generating LTR from seed {seed}"
        raise RuntimeError(msg)

    def set_singles_start(
        self,
        char: str,
        chance: float,
    ):
        self._singles.set_start(char, chance)

    def set_singles_middle(
        self,
        char: str,
        chance: float,
    ):
        self._singles.set_middle(char, chance)

    def set_singles_end(
        self,
        char: str,
        chance: float,
    ):
        self._singles.set_end(char, chance)

    def set_doubles_start(
        self,
        previous1: str,
        char: str,
        chance: float,
    ):
        self._doubles[LTR._CHAR_INDEX[previous1]].set_start(char, chance)

    def set_doubles_middle(
        self,
        previous1: str,
        char: str,
        chance: float,
    ):
        self._doubles[LTR._CHAR_INDEX[previous1]].set_middle(char, chance)

    def set_doubles_end(
        self,
        previous1: str,
        char: str,
        chance: float,
    ):
        self._doubles[LTR._CHAR_INDEX[previous1]].set_end(char, chance)

    def set_triples_start(
        self,
        previous2: str,
        previous1: str,
        char: str,
        chance: float,
    ):
        self._triples[LTR._CHAR_INDEX[previous2]][LTR._CHAR_INDEX[previous1]].set_start(char, chance)

    def set_triples_middle(
        self,
        previous2: str,
        previous1: str,
        char: str,
        chance: float,
    ):
        self._triples[LTR._CHAR_INDEX[previous2]][LTR._CHAR_INDEX[previous1]].set_middle(char, chance)

    def set_triples_end(
        self,
        previous2: str,
        previous1: str,
        char: str,
        chance: float,
    ):
        self._triples[LTR._CHAR_INDEX[previous2]][LTR._CHAR_INDEX[previous1]].set_end(char, chance)


class LTRBlock(ComparableMixin):
    """Stores probability values for characters appearing in different name positions.

        Each LTRBlock contains three probability arrays (start, middle, end) that define
        the likelihood of each character appearing at the start, middle, or end of a name
        segment. These probabilities are cumulative (values increase monotonically) and
        are used with random number generation to select characters.

        References:
        ----------
        Based on swkotor.exe LTR structure:
        - Probability block layout (start/middle/end)
          * Start probabilities: 28 floats (offset 0x09-0x70)
          * Middle probabilities: 28 floats (offset 0x71-0xD8)
          * End probabilities: 28 floats (offset 0xD9-0x140)
        - Binary format (per LetterSet)
          * Start probabilities: num_characters * 4 bytes (float32 each)
          * Middle probabilities: num_characters * 4 bytes (float32 each)
          * End probabilities: num_characters * 4 bytes (float32 each)
        - Original BioWare engine binaries (TODO: Verify with REVA when available)
        https://github.com/mtijanic/nwn-misc/blob/master/nwnltr.c - Original C reference implementation

    Derivations and Other Implementations:
    -------------------------------------
        - https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/LTRObject.ts:19-210
        - https://github.com/th3w1zard1/KotOR_IO/tree/master/KotOR_IO/File
        - https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Resources/KotorLTR/LTR.cs
        - https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Resources/KotorLTR/LTRDecompiler.cs

        Attributes:
        ----------
            _start: Probability array for characters at start of name segment
                Array of 28 floats (one per character)
                Cumulative probabilities (monotonically increasing)
                Used when generating first character or after name breaks

            _middle: Probability array for characters in middle of name segment
                Array of 28 floats (one per character)
                Cumulative probabilities (monotonically increasing)
                Used when generating characters after start but before end

            _end: Probability array for characters at end of name segment
                Array of 28 floats (one per character)
                Cumulative probabilities (monotonically increasing)
                Used when generating final character of name
    """

    COMPARABLE_SEQUENCE_FIELDS = ("_start", "_middle", "_end")

    def __init__(
        self,
        num_characters: int,
    ):
        # Probability array for characters at start of name segment
        # Cumulative probabilities (monotonically increasing)
        self._start: list[float] = [0.0] * num_characters

        # Probability array for characters in middle of name segment
        # Cumulative probabilities (monotonically increasing)
        self._middle: list[float] = [0.0] * num_characters

        # Probability array for characters at end of name segment
        # Cumulative probabilities (monotonically increasing)
        self._end: list[float] = [0.0] * num_characters

    def __eq__(self, other):
        if not isinstance(other, LTRBlock):
            return NotImplemented  # type: ignore[no-any-return]
        return self._start == other._start and self._middle == other._middle and self._end == other._end

    def __hash__(self):
        return hash((tuple(self._start), tuple(self._middle), tuple(self._end)))

    def set_start(
        self,
        char: str,
        chance: float,
    ):
        """Sets the chance of a specified character at the start of the block.

        Args:
        ----
            char: The specified character from LTR.CHARACTER_SET.
            chance: The chance of occurrence from 0.0 to 1.0 inclusive.

        Raises:
        ------
            ValueError: If length of char is not 0 OR if the chance was not between 0.0 and 1.0 inclusive.
            IndexError: If char is not present in LTR.CHARACTER_SET.
        """
        if len(char) != 1:
            msg = "The character specified was not a real character."
            raise ValueError(msg)
        if char not in LTR.CHARACTER_SET:
            msg = "The character specified was invalid."
            raise IndexError(msg)
        if 0.0 > chance > 1.0:
            msg = "The chance specified must be between 0.0 and 1.0 inclusive."
            raise ValueError(msg)

        char_id = LTR._CHAR_INDEX[char]
        self._start[char_id] = chance

    def set_middle(
        self,
        char: str,
        chance: float,
    ):
        """Sets the chance of a specified character at the middle of the block.

        Args:
        ----
            char: char: The specified character from LTR.CHARACTER_SET.
            chance: The chance of occurrence from 0.0 to 1.0 inclusive.

        Raises:
        ------
            ValueError: If length of char is not 0 OR if the chance was not between 0.0 and 1.0 inclusive.
            IndexError: If char is not present in LTR.CHARACTER_SET.
        """
        if len(char) != 1:
            msg = "The character specified was not a real character."
            raise ValueError(msg)
        if char not in LTR.CHARACTER_SET:
            msg = "The character specified was invalid."
            raise IndexError(msg)
        if 0.0 > chance > 1.0:
            msg = "The chance specified must be between 0.0 and 1.0 inclusive."
            raise ValueError(msg)

        char_id = LTR._CHAR_INDEX[char]
        self._middle[char_id] = chance

    def set_end(
        self,
        char: str,
        chance: float,
    ):
        """Sets the chance of a specified character at the end of the block.

        Args:
        ----
            char: char: The specified character from LTR.CHARACTER_SET.
            chance: The chance of occurrence from 0.0 to 1.0 inclusive.

        Raises:
        ------
            ValueError: If length of char is not 0 OR if the chance was not between 0.0 and 1.0 inclusive.
            IndexError: If char is not present in LTR.CHARACTER_SET.
        """
        if len(char) != 1:
            msg = "The character specified was not a real character."
            raise ValueError(msg)
        if char not in LTR.CHARACTER_SET:
            msg = "The character specified was invalid."
            raise IndexError(msg)
        if 0.0 > chance > 1.0:
            msg = "The chance specified must be between 0.0 and 1.0 inclusive."
            raise ValueError(msg)

        char_id = LTR._CHAR_INDEX[char]
        self._end[char_id] = chance

    def get_start(
        self,
        char: str,
    ) -> float:
        """Returns the chance of a specified character at the start of the block.

        Args:
        ----
            char: char: The specified character from LTR.CHARACTER_SET.

        Raises:
        ------
            ValueError: If length of char is not 0.
            IndexError: If char is not present in LTR.CHARACTER_SET.

        Returns:
        -------
            The chance of a specified character at the start of the block.

        """
        if len(char) != 1:
            msg = "The character specified was not a real character."
            raise ValueError(msg)
        if char not in LTR.CHARACTER_SET:
            msg = "The character specified was invalid."
            raise IndexError(msg)

        char_id = LTR._CHAR_INDEX[char]
        return self._start[char_id]

    def get_middle(
        self,
        char: str,
    ) -> float:
        """Returns the chance of a specified character at the middle of the block.

        Args:
        ----
            char: char: The specified character from LTR.CHARACTER_SET.

        Raises:
        ------
            ValueError: If length of char is not 0.
            IndexError: If char is not present in LTR.CHARACTER_SET.

        Returns:
        -------
            The chance of a specified character at the start of the block.

        """
        if len(char) != 1:
            msg = "The character specified was not a real character."
            raise ValueError(msg)
        if char not in LTR.CHARACTER_SET:
            msg = "The character specified was invalid."
            raise IndexError(msg)

        char_id = LTR._CHAR_INDEX[char]
        return self._middle[char_id]

    def get_end(
        self,
        char: str,
    ) -> float:
        """Returns the chance of a specified character at the end of the block.

        Args:
        ----
            char: char: The specified character from LTR.CHARACTER_SET.

        Raises:
        ------
            ValueError: If length of char is not 0.
            IndexError: If char is not present in LTR.CHARACTER_SET.

        Returns:
        -------
            The chance of a specified character at the start of the block.

        """
        if len(char) != 1:
            msg = "The character specified was not a real character."
            raise ValueError(msg)
        if char not in LTR.CHARACTER_SET:
            msg = "The character specified was invalid."
            raise IndexError(msg)

        char_id = LTR._CHAR_INDEX[char]
        return self._end[char_id]



"""LYT (layout) package: room/track/obstacle/door layout and ASCII I/O for modules."""

from __future__ import annotations

from pykotor.resource.formats.lyt.lyt_data import (
    LYT,
    LYTRoom,
    LYTTrack,
    LYTObstacle,
    LYTDoorHook,
)
from pykotor.resource.formats.lyt.io_lyt import (
    LYTAsciiReader,
    LYTAsciiWriter,
)
from pykotor.resource.formats.lyt.lyt_auto import write_lyt, read_lyt, bytes_lyt

__all__ = [
    "LYT",
    "LYTAsciiReader",
    "LYTAsciiWriter",
    "LYTDoorHook",
    "LYTObstacle",
    "LYTRoom",
    "LYTTrack",
    "bytes_lyt",
    "read_lyt",
    "write_lyt",
]



"""Model manipulation utilities for MDL/MDX files.

This module provides utility functions for manipulating KotOR model geometry at the binary level,
including texture/lightmap iteration, node renaming, and geometry offset calculations.

References:
----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/kotorblender/tree/master/io_scene_kotor/format/mdl/
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/odyssey/OdysseyModel.ts


    Note: Magic float values (_MESH_FP0_K1, etc.) are game-specific floating point offsets for binary patching
"""

from __future__ import annotations

import math
import struct

from collections import deque
from copy import deepcopy
from typing import TYPE_CHECKING, Any, NamedTuple

from pykotor.common.misc import Game
from pykotor.common.stream import BinaryReader
from utility.common.geometry import Vector4

if TYPE_CHECKING:
    from collections.abc import Generator

    from utility.common.geometry import Vector3

_GEOM_ROOT_FP0_K1: int = 4273776
_GEOM_ROOT_FP1_K1: int = 4216096
_GEOM_ROOT_FP0_K2: int = 4285200
_GEOM_ROOT_FP1_K2: int = 4216320

_GEOM_ANIM_FP0_K1: int = 4273392
_GEOM_ANIM_FP1_K1: int = 4451552
_GEOM_ANIM_FP0_K2: int = 4284816
_GEOM_ANIM_FP1_K2: int = 4522928

_NODE_TYPE_MESH: int = 32
_MESH_HEADER_SIZE_K1: int = 332
_MESH_HEADER_SIZE_K2: int = 340
_MESH_FP0_K1: int = 4216656
_MESH_FP1_K1: int = 4216672
_MESH_FP0_K2: int = 4216880
_MESH_FP1_K2: int = 4216896

_NODE_TYPE_SKIN: int = 64
_SKIN_HEADER_SIZE = 108
_SKIN_FP0_K1: int = 4216592
_SKIN_FP1_K1: int = 4216608
_SKIN_FP0_K2: int = 4216816
_SKIN_FP1_K2: int = 4216832

_NODE_TYPE_DANGLY: int = 256
_DANGLY_HEADER_SIZE: int = 28
_DANGLY_FP0_K1: int = 4216640
_DANGLY_FP1_K1: int = 4216624
_DANGLY_FP0_K2: int = 4216864
_DANGLY_FP1_K2: int = 4216848

_NODE_TYPE_SABER: int = 2048
_SABER_HEADER_SIZE: int = 20
_SABER_FP0_K1: int = 4216656
_SABER_FP1_K1: int = 4216672
_SABER_FP0_K2: int = 4216880
_SABER_FP1_K2: int = 4216896

_NODE_TYPE_AABB: int = 512
_AABB_HEADER_SIZE: int = 4
_AABB_FP0_K1: int = 4216656
_AABB_FP1_K1: int = 4216672
_AABB_FP0_K2: int = 4216880
_AABB_FP1_K2: int = 4216896

_NODE_TYPE_LIGHT: int = 2
_LIGHT_HEADER_SIZE: int = 92

_NODE_TYPE_EMITTER: int = 4
_EMITTER_HEADER_SIZE: int = 224


class MDLMDXTuple(NamedTuple):
    mdl: bytes | bytearray
    mdx: bytes | bytearray


def rename(
    data: bytes,
    name: str,
) -> bytes:
    return data[:20] + name.ljust(32, "\0").encode("ascii") + data[52:]


def _traverse_mesh_nodes(reader: BinaryReader) -> Generator[int, Any, None]:
    """Traverse the model tree and yield offsets of mesh nodes.

    Args:
    ----
        reader: BinaryReader positioned at the model data

    Yields:
    ------
        int: Offset of each mesh node in the data
    """
    reader.seek(168)
    root_offset: int = reader.read_uint32()

    nodes: deque[int] = deque([root_offset])
    while nodes:
        node_offset: int = nodes.popleft()
        reader.seek(node_offset)
        node_id: int = reader.read_uint32()

        reader.seek(node_offset + 44)
        child_offsets_offset: int = reader.read_uint32()
        child_offsets_count: int = reader.read_uint32()

        reader.seek(child_offsets_offset)
        nodes.extend(reader.read_uint32() for _ in range(child_offsets_count))

        if node_id & _NODE_TYPE_MESH:
            yield node_offset


def iterate_textures_and_lightmaps(
    data: bytes,
) -> Generator[str, Any, None]:
    """Iterate over all unique texture and lightmap names in the model.

    Args:
    ----
        data: The model data bytes

    Yields:
    ------
        str: Unique texture or lightmap names (lowercased)
    """
    seen_names: set[str] = set()

    with BinaryReader.from_bytes(data, 12) as reader:
        for node_offset in _traverse_mesh_nodes(reader):
            # Extract texture name
            reader.seek(node_offset + 168)
            name: str = reader.read_string(32, encoding="ascii", errors="ignore").strip().lower()
            if name and name != "null" and name not in seen_names and name != "dirt":
                seen_names.add(name)
                yield name

            # Extract lightmap name
            reader.seek(node_offset + 200)
            name = reader.read_string(32, encoding="ascii", errors="ignore").strip().lower()
            if name and name != "null" and name not in seen_names:
                seen_names.add(name)
                yield name


def iterate_textures(
    data: bytes,
) -> Generator[str, Any, None]:
    """Iterate over all unique texture names in the model.

    Args:
    ----
        data: The model data bytes

    Yields:
    ------
        str: Unique texture names (lowercased)
    """
    texture_caseset: set[str] = set()

    with BinaryReader.from_bytes(data, 12) as reader:
        for node_offset in _traverse_mesh_nodes(reader):
            reader.seek(node_offset + 168)
            texture = reader.read_string(32, encoding="ascii", errors="ignore").strip()
            if (
                texture
                and texture != "NULL"
                and texture.lower() not in texture_caseset
                and texture.lower() != "dirt"
            ):
                texture_caseset.add(texture.lower())
                yield texture.lower()


def iterate_lightmaps(
    data: bytes | bytearray,
) -> Generator[str, Any, None]:
    """Iterate over all unique lightmap names in the model.

    Args:
    ----
        data: The model data bytes

    Yields:
    ------
        str: Unique lightmap names
    """
    lightmaps_caseset: set[str] = set()
    with BinaryReader.from_bytes(data, 12) as reader:
        for node_offset in _traverse_mesh_nodes(reader):
            reader.seek(node_offset + 200)
            lightmap: str = reader.read_string(32, encoding="ascii", errors="ignore").strip()
            lowercase_lightmap: str = lightmap.lower()
            if lightmap and lightmap != "NULL" and lowercase_lightmap not in lightmaps_caseset:
                lightmaps_caseset.add(lowercase_lightmap)
                yield lightmap


def change_textures(
    data: bytes | bytearray,
    textures: dict[str, str],
) -> bytes | bytearray:
    return _change_mesh_string_field(data, textures, field_offset=168)


def change_lightmaps(
    data: bytes | bytearray,
    textures: dict[str, str],
) -> bytes | bytearray:
    return _change_mesh_string_field(data, textures, field_offset=200)


def _change_mesh_string_field(
    data: bytes | bytearray,
    replacements: dict[str, str],
    *,
    field_offset: int,
) -> bytes:
    """Replace a 32-byte mesh string field (texture/lightmap) in all mesh nodes."""
    parsed_data: bytearray = bytearray(data)
    offsets: dict[str, list[int]] = {}

    normalized_replacements: dict[str, str] = {
        old_value.lower(): new_value.lower() for old_value, new_value in replacements.items()
    }

    with BinaryReader.from_bytes(parsed_data, 12) as reader:
        reader.seek(168)
        root_offset: int = reader.read_uint32()

        nodes: list[int] = [root_offset]
        while nodes:
            node_offset: int = nodes.pop()
            reader.seek(node_offset)
            node_id: int = reader.read_uint32()

            reader.seek(node_offset + 44)
            child_offsets_offset: int = reader.read_uint32()
            child_offsets_count: int = reader.read_uint32()

            reader.seek(child_offsets_offset)
            nodes.extend(reader.read_uint32() for _i in range(child_offsets_count))
            if node_id & 32:
                reader.seek(node_offset + field_offset)
                field_value: str = reader.read_string(32).lower()

                if field_value in normalized_replacements:
                    if field_value in offsets:
                        offsets[field_value].append(node_offset + field_offset)
                    else:
                        offsets[field_value] = [node_offset + field_offset]

        for original_value, offsets_list in offsets.items():
            for offset in offsets_list:
                actual_offset: int = offset + 12
                parsed_data = (
                    parsed_data[:actual_offset]
                    + struct.pack(
                        "32s",
                        normalized_replacements[original_value].ljust(32, "\0").encode("ascii"),
                    )
                    + parsed_data[actual_offset + 32 :]
                )

    return bytes(parsed_data)


def detect_version(
    data: bytes | bytearray,
) -> Game:
    pointer: int = struct.unpack("I", data[12:16])[0]
    return Game.K1 if pointer == _GEOM_ROOT_FP0_K1 else Game.K2


def convert_to_k1(
    data: bytes | bytearray,
) -> bytes | bytearray:
    if detect_version(data) == Game.K1:
        return data

    trim: list[tuple[int, int]] = []

    with BinaryReader.from_bytes(data, 12) as reader:
        reader.seek(168)
        root_offset: int = reader.read_uint32()

        nodes: list[int] = [root_offset]
        while nodes:
            node_offset: int = nodes.pop()
            reader.seek(node_offset)
            node_type: int = reader.read_uint16()

            if node_type & 32:
                trim.append((node_type, node_offset))

            reader.seek(node_offset + 44)
            child_offsets_offset: int = reader.read_uint32()
            child_offsets_count: int = reader.read_uint32()

            reader.seek(child_offsets_offset)
            nodes.extend(reader.read_uint32() for _ in range(child_offsets_count))
    start: bytes | bytearray = data[:12]
    parsed_data: bytearray = bytearray(data[12:])

    parsed_data[:4] = struct.pack("I", _GEOM_ROOT_FP0_K1)
    parsed_data[4:8] = struct.pack("I", _GEOM_ROOT_FP1_K1)

    # TODO(NickHugi): Animations
    for node_type, node_offset in trim:
        mesh_start: int = node_offset + 80  # Start of mesh header

        offset_start: int = node_offset + 80 + 332  # Location of start of bytes to be shifted
        offset_size: int = 8  # How many bytes we have to shift

        if node_type & _NODE_TYPE_SKIN:
            offset_size += _SKIN_HEADER_SIZE
            parsed_data[mesh_start : mesh_start + 4] = struct.pack("I", _MESH_FP0_K1)
            parsed_data[mesh_start + 4 : mesh_start + 8] = struct.pack("I", _MESH_FP1_K1)

        if node_type & _NODE_TYPE_DANGLY:
            offset_size += _DANGLY_HEADER_SIZE
            parsed_data[mesh_start : mesh_start + 4] = struct.pack("I", _DANGLY_FP0_K1)
            parsed_data[mesh_start + 4 : mesh_start + 8] = struct.pack("I", _DANGLY_FP1_K1)

        if node_type & _NODE_TYPE_SABER:
            offset_size += _SABER_HEADER_SIZE
            parsed_data[mesh_start : mesh_start + 4] = struct.pack("I", _SABER_FP0_K1)
            parsed_data[mesh_start + 4 : mesh_start + 8] = struct.pack("I", _SABER_FP1_K1)

        if node_type & _NODE_TYPE_AABB:
            offset_size += _AABB_HEADER_SIZE
            parsed_data[mesh_start : mesh_start + 4] = struct.pack("I", _AABB_FP0_K1)
            parsed_data[mesh_start + 4 : mesh_start + 8] = struct.pack("I", _AABB_FP1_K1)

        shifting: bytearray = parsed_data[offset_start : offset_start + offset_size]
        parsed_data[offset_start - 8 : offset_start - 8 + offset_size] = shifting

    return bytes(start + parsed_data)


def convert_to_k2(  # noqa: C901, PLR0915, PLR0912
    data: bytes | bytearray,
) -> bytes | bytearray:
    if detect_version(data) == Game.K2:
        return data

    offsets: dict[int, int] = {}  # Maps the offset for an offset to its offset
    mesh_offsets: list[list[int]] = []  # tuple of (Offset to every mesh node, Node type)
    anim_offsets: list[int] = []

    # First, we build a dictionary of every offset in the file plus a list of the mesh nodes
    with BinaryReader.from_bytes(data, 12) as reader:

        def node_recursive(  # noqa: C901, PLR0912, PLR0915
            offset_to_root_offset: int,
        ) -> None:
            nodes: list[int] = [offset_to_root_offset]
            while nodes:
                offset_to_node_offset: int = nodes.pop()
                reader.seek(offset_to_node_offset)
                node_offset: int = reader.read_uint32()
                offsets[offset_to_node_offset] = node_offset  # Geometry header/Node children offsets array -> Node offset

                reader.seek(node_offset)
                node_type: int = reader.read_uint16()

                base_offset: int = node_offset + 80
                if node_type & _NODE_TYPE_MESH:
                    mesh_offsets.append([node_offset, node_type])

                    reader.seek(base_offset + 8)
                    offsets[base_offset + 8] = reader.read_uint32()  # Node header -> Face array offset

                    reader.seek(base_offset + 176)
                    offsets[base_offset + 176] = reader.read_uint32()  # Node header -> Vertex indices count array
                    indices_array_count: int = reader.read_uint32()

                    reader.seek(base_offset + 188)
                    offsets[base_offset + 188] = reader.read_uint32()  # Node header -> Vertex indices locations array
                    if indices_array_count == 1:
                        indices_locations_offset: int = offsets[base_offset + 188]
                        reader.seek(indices_locations_offset)
                        offsets[indices_locations_offset] = reader.read_uint32()  # Vertex indices locations array -> Vertex indices array

                    reader.seek(base_offset + 200)
                    offsets[base_offset + 200] = reader.read_uint32()  # Node header -> Inverted counter array

                    reader.seek(base_offset + 328)
                    offsets[base_offset + 328] = reader.read_uint32()  # Node header -> Vertex array

                    base_offset += _MESH_HEADER_SIZE_K1

                if node_type & _NODE_TYPE_LIGHT:
                    reader.seek(base_offset + 4)
                    offsets[base_offset + 4] = reader.read_uint32()  # Node header -> Lens flare size array

                    reader.seek(base_offset + 16)
                    offsets[base_offset + 16] = reader.read_uint32()  # Node header -> Lens flare size array

                    reader.seek(base_offset + 28)
                    offsets[base_offset + 28] = reader.read_uint32()  # Node header -> Lens flare positions array

                    reader.seek(base_offset + 40)
                    offsets[base_offset + 40] = reader.read_uint32()  # Node header -> Flare colour shifts array

                    reader.seek(base_offset + 52)
                    offsets[base_offset + 52] = reader.read_uint32()  # Node header -> Flare texture names offsets array
                    texture_count: int = reader.read_uint32()

                    for i in range(texture_count):
                        reader.seek(offsets[base_offset + 52] + i * 4)
                        offsets[offsets[base_offset + 52] + i * 4] = reader.read_uint32()

                if node_type & _NODE_TYPE_EMITTER:
                    ...  # No offsets. Hooray.

                if node_type & _NODE_TYPE_SKIN:
                    reader.seek(base_offset + 20)
                    offsets[base_offset + 20] = reader.read_uint32()  # Node header -> Bone map array

                    reader.seek(base_offset + 28)
                    offsets[base_offset + 28] = reader.read_uint32()  # Node header -> QBones array

                    reader.seek(base_offset + 40)
                    offsets[base_offset + 40] = reader.read_uint32()  # Node header -> TBones Array

                    reader.seek(base_offset + 52)
                    offsets[base_offset + 52] = reader.read_uint32()  # Node header -> Array8

                    base_offset += _SKIN_HEADER_SIZE

                if node_type & _NODE_TYPE_DANGLY:
                    reader.seek(base_offset + 0)
                    offsets[base_offset + 0] = reader.read_uint32()  # Node header -> Dangly constraint array

                    reader.seek(base_offset + 24)
                    offsets[base_offset + 24] = reader.read_uint32()  # Node header -> Unknown

                    base_offset += _DANGLY_HEADER_SIZE

                if node_type & _NODE_TYPE_AABB:
                    reader.seek(base_offset + 0)
                    offsets[base_offset + 0] = reader.read_uint32()  # Node header -> AABB root node

                    aabbs: list[int] = [offsets[base_offset + 0]]
                    while aabbs:
                        aabb: int = aabbs.pop()

                        reader.seek(aabb + 24)
                        leaf0: int = reader.read_uint32()
                        if leaf0:
                            aabbs.append(leaf0)
                            offsets[aabb + 24] = leaf0  # AABB Node -> AABB child node

                        reader.seek(aabb + 28)
                        leaf1: int = reader.read_uint32()
                        if leaf1:
                            aabbs.append(leaf1)
                            offsets[aabb + 28] = leaf1  # AABB Node -> AABB child node

                    base_offset += _AABB_HEADER_SIZE

                if node_type & _NODE_TYPE_SABER:
                    reader.seek(base_offset + 0)
                    offsets[base_offset + 0] = reader.read_uint32()  # Node header -> Saber Verts array

                    reader.seek(base_offset + 4)
                    offsets[base_offset + 4] = reader.read_uint32()  # Node header -> Saber UVs array

                    reader.seek(base_offset + 8)
                    offsets[base_offset + 8] = reader.read_uint32()  # Node header -> Saber Normals array

                    base_offset += _SABER_HEADER_SIZE

                reader.seek(node_offset + 8)
                offsets[node_offset + 8] = reader.read_uint32()  # Node header -> Geometry header

                reader.seek(node_offset + 12)
                offsets[node_offset + 12] = reader.read_uint32()  # Node header -> Parent node

                reader.seek(node_offset + 56)
                offsets[node_offset + 56] = reader.read_uint32()  # Node header -> controller array offset

                reader.seek(node_offset + 68)
                offsets[node_offset + 68] = reader.read_uint32()  # Node header -> controller data offset

                reader.seek(node_offset + 44)
                child_offsets_offset: int = reader.read_uint32()
                child_offsets_count: int = reader.read_uint32()
                offsets[node_offset + 44] = child_offsets_offset  # Node header -> Child node offsets array

                nodes = [child_offsets_offset + i * 4 for i in range(child_offsets_count)]
                nodes.insert(0, offset_to_root_offset)

        reader.seek(88)
        anim_locations_offset: int = reader.read_uint32()
        anim_count: int = reader.read_uint32()

        reader.seek(168)
        reader.read_uint32()
        reader.read_uint32()

        reader.seek(184)
        reader.read_uint32()
        name_count: int = reader.read_uint32()

        reader.seek(40)
        offsets[40] = reader.read_uint32()  # Model header -> Root node

        reader.seek(88)
        offsets[88] = reader.read_uint32()  # Model header -> Animation array

        reader.seek(168)
        offsets[168] = reader.read_uint32()  # Model header -> Head root

        reader.seek(184)
        offsets[184] = reader.read_uint32()  # Model header -> Name offsets array

        for i in range(name_count):
            reader.seek(offsets[184] + i * 4)
            offsets[offsets[184] + i * 4] = reader.read_uint32()  # Name offset array -> Name

        for i in range(anim_count):
            reader.seek(anim_locations_offset + i * 4)
            animation_offset: int = reader.read_uint32()
            offsets[anim_locations_offset + i * 4] = animation_offset  # Event locations array -> Offset to event
            anim_offsets.append(animation_offset)

            reader.seek(animation_offset + 120)
            offsets[animation_offset + 120] = reader.read_uint32()  # Animation header -> Offset to event array

            reader.seek(animation_offset + 40)
            node_recursive(animation_offset + 40)

        node_recursive(168)

    # Second, we will update the function pointers to use K2 values instead of K1
    mdx_size: bytes | bytearray = data[8:12]
    parsed_data: bytearray = bytearray(data[12:])

    parsed_data[:4] = struct.pack("I", _GEOM_ROOT_FP0_K2)
    parsed_data[4:8] = struct.pack("I", _GEOM_ROOT_FP1_K2)

    for anim_offset in anim_offsets:
        parsed_data[anim_offset : anim_offset + 4] = struct.pack("I", _GEOM_ANIM_FP0_K2)
        parsed_data[anim_offset + 4 : anim_offset + 8] = struct.pack("I", _GEOM_ANIM_FP1_K2)

    for node_offset, node_type in mesh_offsets:
        mesh_start: int = node_offset + 80  # Start of mesh header

        if node_type & _NODE_TYPE_SKIN:
            parsed_data[mesh_start : mesh_start + 4] = struct.pack("I", _MESH_FP0_K2)
            parsed_data[mesh_start + 4 : mesh_start + 8] = struct.pack("I", _MESH_FP1_K2)

        if node_type & _NODE_TYPE_DANGLY:
            parsed_data[mesh_start : mesh_start + 4] = struct.pack("I", _DANGLY_FP0_K2)
            parsed_data[mesh_start + 4 : mesh_start + 8] = struct.pack("I", _DANGLY_FP1_K2)

        if node_type & _NODE_TYPE_SABER:
            parsed_data[mesh_start : mesh_start + 4] = struct.pack("I", _SABER_FP0_K1)
            parsed_data[mesh_start + 4 : mesh_start + 8] = struct.pack("I", _SABER_FP1_K2)

        if node_type & _NODE_TYPE_AABB:
            parsed_data[mesh_start : mesh_start + 4] = struct.pack("I", _AABB_FP0_K2)
            parsed_data[mesh_start + 4 : mesh_start + 8] = struct.pack("I", _AABB_FP1_K2)

    offsets = dict(sorted(offsets.items(), reverse=True))

    # Third, we will add the extra data in the mesh headers and update all offsets in our dictionary
    shifted: int = 0
    for i in range(len(mesh_offsets)):
        node_offset, node_type = mesh_offsets[i]
        insert_location: int = node_offset + 80 + 324
        parsed_data = bytearray(parsed_data[:insert_location] + bytes([0] * 8) + parsed_data[insert_location:])

        for offset_location, offset_value in deepcopy(offsets).items():
            if insert_location < offset_location:
                del offsets[offset_location]
                if offset_location + 8 in offsets:
                    msg = "whoops"
                    raise ValueError(msg)
                offsets[offset_location + 8] = offset_value

        for offset_location, offset_value in deepcopy(offsets).items():
            if insert_location < offset_value:
                offsets[offset_location] = offset_value + 8

        for j in range(len(mesh_offsets)):
            if insert_location < mesh_offsets[j][0]:
                mesh_offsets[j][0] += 8

        shifted += 8

    # Finally, we update the offsets in the bytes with the offsets in our dictionary
    for offset_location, offset_value in offsets.items():
        parsed_data[offset_location : offset_location + 4] = struct.pack("I", offset_value)

    return struct.pack("I", 0) + struct.pack("I", len(parsed_data)) + mdx_size + parsed_data


def transform(
    data: bytes | bytearray,
    translation: Vector3,
    rotation: float,
) -> bytes | bytearray:
    orientation: Vector4 = Vector4.from_euler(0, 0, math.radians(rotation))
    mdx_size: int = struct.unpack("I", data[8:12])[0]
    parsed_data: bytearray = bytearray(data[12:])

    with BinaryReader.from_bytes(parsed_data) as reader:
        reader.seek(44)
        node_count: int = reader.read_uint32()

        reader.seek(168)
        root_offset: int = reader.read_uint32()

        reader.seek(root_offset)
        reader.read_uint16()
        reader.read_uint16()
        reader.read_uint32()
        reader.skip(6)
        reader.skip(4)
        reader.skip(4)
        reader.skip(4 * 3)
        reader.skip(4 * 4)

        reader.seek(root_offset + 44)
        child_array_offset: int = reader.read_uint32()
        child_count: int = reader.read_uint32()

    if child_count == 0:
        return parsed_data

    root_child_array_offset: int = len(parsed_data)
    insert_node_offset: int = len(parsed_data) + 4
    insert_controller_offset: int = insert_node_offset + 80
    insert_controller_data_offset: int = insert_controller_offset + 32

    # Increase global node count by 1
    parsed_data[44:48] = struct.pack("I", node_count + 1)

    # Update the offset the array of child offsets to our injected array
    parsed_data[root_offset + 44 : root_offset + 48] = struct.pack(
        "I",
        root_child_array_offset,
    )
    # Set the root node to have 1 child
    parsed_data[root_offset + 48 : root_offset + 52] = struct.pack("I", 1)
    parsed_data[root_offset + 52 : root_offset + 56] = struct.pack("I", 1)

    # Populate the injected new root child offsets array
    # It will contain our new node
    parsed_data += struct.pack("I", insert_node_offset)
    # Create the new node
    parsed_data += struct.pack(
        "HHHH II fff ffff III III III",
        1,  # Node Type
        node_count + 1,  # Node ID
        1,  # Label ID (steal some existing node's label)
        0,  # Padding
        0,
        root_offset,
        *translation,  # Node Position
        orientation.w,
        orientation.x,
        orientation.y,
        orientation.z,  # Node Orientation
        child_array_offset,  # Child Array Offset
        child_count,
        child_count,
        insert_controller_offset,  # Controller Array
        2,
        2,
        insert_controller_data_offset,  # Controller Data Array
        9,
        9,
    )
    # Inject controller and controller data of new node to the end of the file
    parsed_data += struct.pack("IHHHHBBBB", 8, 0xFFFF, 1, 0, 1, 3, 0, 0, 0)
    parsed_data += struct.pack("IHHHHBBBB", 20, 0xFFFF, 1, 4, 5, 4, 0, 0, 0)
    parsed_data += struct.pack("ffff", 0.0, *translation)
    parsed_data += struct.pack("fffff", 0.0, *orientation)

    return struct.pack("III", 0, len(parsed_data), mdx_size) + parsed_data


def flip(  # noqa: C901, PLR0912, PLR0915
    mdl_data: bytes | bytearray,
    mdx_data: bytes | bytearray,
    *,
    flip_x: bool,
    flip_y: bool,
) -> MDLMDXTuple:
    # If neither bools are set to True, no transformations need to be done and we can just return the original data
    if not flip_x and not flip_y:
        return MDLMDXTuple(mdl_data, mdx_data)

    # The data we need to change:
    #    1. The vertices stored in the MDL
    #    2. The vertex positions, normals, stored in the MDX

    # Trim the data to correct the offsets
    mdl_start: bytearray = bytearray(mdl_data)[:12]
    parsed_mdl_data = bytearray(mdl_data)[12:]
    parsed_mdx_data = bytearray(mdx_data)

    mdl_vertex_offsets: list[tuple[int, int]] = []  # This is a list of tuples: (count, offset)
    mdx_vertex_offsets: list[tuple[int, int, int, int]] = []  # This is a list of tuples: (count, offset, stride, position)
    mdx_normal_offsets: list[tuple[int, int, int, int]] = []  # This is a list of tuples: (count, offset, stride, position)
    elements_offsets: list[tuple[int, int]] = []  # This is a list of tuples: (count, offset)
    faces_offsets: list[tuple[int, int]] = []  # This is a list of tuples: (count, offset)
    with BinaryReader.from_bytes(parsed_mdl_data) as reader:
        reader.seek(168)
        root_offset: int = reader.read_uint32()

        nodes: list[int] = [root_offset]
        while nodes:
            node_offset: int = nodes.pop()
            reader.seek(node_offset)
            node_id: int = reader.read_uint32()

            mdl_vertex_offsets.append((1, node_offset + 16))

            # Need to determine the location of the position controller
            reader.seek(node_offset + 56)
            controllers_offset: int = reader.read_uint32()
            controllers_count: int = reader.read_uint32()

            reader.seek(node_offset + 68)
            controller_datas_offset: int = reader.read_uint32()
            reader.read_uint32()

            for i in range(controllers_count):
                reader.seek(controllers_offset + i * 16)
                controller_type: int = reader.read_uint32()
                if controller_type == 8:  # noqa: PLR2004
                    reader.skip(6)
                    data_offset: int = reader.read_uint16()
                    mdl_vertex_offsets.append(
                        (1, controller_datas_offset + data_offset * 4),
                    )

            reader.seek(node_offset + 44)
            child_offsets_offset: int = reader.read_uint32()
            child_offsets_count: int = reader.read_uint32()

            reader.seek(child_offsets_offset)
            nodes.extend(reader.read_uint32() for i in range(child_offsets_count))
            if node_id & 32:
                reader.seek(node_offset + 80)
                fp: int = reader.read_uint32()
                tsl: bool = fp not in {
                    _MESH_FP0_K1,
                    _SKIN_FP0_K1,
                    _DANGLY_FP0_K2,
                    _AABB_FP0_K1,
                    _SABER_FP0_K1,
                }

                reader.seek(node_offset + 80 + 8)
                faces_offset: int = reader.read_uint32()
                faces_count: int = reader.read_uint32()
                faces_offsets.append((faces_count, faces_offset))

                reader.seek(node_offset + 80 + 188)
                offset_to_elements_offset = reader.read_uint32()
                reader.seek(offset_to_elements_offset)
                elements_offset: int = reader.read_uint32()
                elements_offsets.append((faces_count, elements_offset))

                reader.seek(node_offset + 80 + 304)
                vertex_count: int = reader.read_uint16()
                reader.seek(node_offset + 80 + 336 if tsl else node_offset + 80 + 328)
                vertex_offset: int = reader.read_uint32()
                mdl_vertex_offsets.append((vertex_count, vertex_offset))

                reader.seek(node_offset + 80 + 252)
                mdx_stride: int = reader.read_uint32()
                reader.read_uint32()
                reader.seek(node_offset + 80 + 260)
                mdx_offset_pos: int = reader.read_uint32()
                mdx_offset_norm: int = reader.read_uint32()
                reader.seek(node_offset + 80 + 332 if tsl else node_offset + 80 + 324)
                mdx_start: int = reader.read_uint32()
                mdx_vertex_offsets.append(
                    (vertex_count, mdx_start, mdx_stride, mdx_offset_pos),
                )
                mdx_normal_offsets.append(
                    (vertex_count, mdx_start, mdx_stride, mdx_offset_norm),
                )

    # Fix vertex order
    if flip_x != flip_y:
        for count, start_offset in elements_offsets:
            for i in range(count):
                offset: int = start_offset + i * 6
                v1: int = struct.unpack("H", parsed_mdl_data[offset : offset + 2])[0]
                v2: int = struct.unpack("H", parsed_mdl_data[offset + 2 : offset + 4])[0]
                v3: int = struct.unpack("H", parsed_mdl_data[offset + 4 : offset + 6])[0]
                parsed_mdl_data[offset : offset + 2] = struct.pack("H", v1)
                parsed_mdl_data[offset + 2 : offset + 4] = struct.pack("H", v3)
                parsed_mdl_data[offset + 4 : offset + 6] = struct.pack("H", v2)
        for count, start_offset in faces_offsets:
            for i in range(count):
                offset = start_offset + i * 32 + 26
                v1 = struct.unpack("H", parsed_mdl_data[offset : offset + 2])[0]
                v2 = struct.unpack("H", parsed_mdl_data[offset + 2 : offset + 4])[0]
                v3 = struct.unpack("H", parsed_mdl_data[offset + 4 : offset + 6])[0]
                parsed_mdl_data[offset : offset + 2] = struct.pack("H", v1)
                parsed_mdl_data[offset + 2 : offset + 4] = struct.pack("H", v3)
                parsed_mdl_data[offset + 4 : offset + 6] = struct.pack("H", v2)

    # Update the MDL vertices
    for count, start_offset in mdl_vertex_offsets:
        for i in range(count):
            offset = start_offset + i * 12
            if flip_x:
                x: float = struct.unpack("f", parsed_mdl_data[offset : offset + 4])[0]
                parsed_mdl_data[offset : offset + 4] = struct.pack("f", -x)
            if flip_y:
                y: float = struct.unpack("f", parsed_mdl_data[offset + 4 : offset + 8])[0]
                parsed_mdl_data[offset + 4 : offset + 8] = struct.pack("f", -y)

    # Update the MDX vertices
    for count, start_offset, stride, position in mdx_vertex_offsets:
        for i in range(count):
            offset = start_offset + i * stride + position
            if flip_x:
                x = struct.unpack("f", parsed_mdx_data[offset : offset + 4])[0]
                parsed_mdx_data[offset : offset + 4] = struct.pack("f", -x)
            if flip_y:
                y = struct.unpack("f", parsed_mdx_data[offset + 4 : offset + 8])[0]
                parsed_mdx_data[offset + 4 : offset + 8] = struct.pack("f", -y)

    # Update the MDX normals
    for count, start_offset, stride, position in mdx_normal_offsets:
        for i in range(count):
            offset = start_offset + i * stride + position
            if flip_x:
                x = struct.unpack("f", parsed_mdx_data[offset : offset + 4])[0]
                parsed_mdx_data[offset : offset + 4] = struct.pack("f", -x)
            if flip_y:
                y = struct.unpack("f", parsed_mdx_data[offset + 4 : offset + 8])[0]
                parsed_mdx_data[offset + 4 : offset + 8] = struct.pack("f", -y)

    # Re-add the first 12 bytes
    parsed_mdl_data = mdl_start + parsed_mdl_data

    return MDLMDXTuple(parsed_mdl_data, parsed_mdx_data)



"""Utilities for authoring valid KotOR installation layouts.

This module provides a shared write-side backend for synthetic or curated
installations used by tests, CLI workflows, and tooling. The generated output is
structurally valid and readable by ``Installation``.

CLI usage::

    pykotor create-installation /tmp/my_kotor --game k1
    pykotor create-installation /tmp/my_tsl --game k2

Programmatic usage::

    from pathlib import Path
    from pykotor.common.misc import Game
    from pykotor.tools.create_installation import create_installation

    create_installation(
        Path("/tmp/my_kotor"),
        Game.K1,
        bif_resources={"party.bif": {"c_bantha.utc": b"..."}},
        override_resources={"my_script.ncs": b"..."},
    )
"""

from __future__ import annotations

import logging

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from pykotor.common.misc import Game

log = logging.getLogger(__name__)

_DEFAULT_TLK_TEXT = "ERROR: FATAL COMPILER ERROR"


def _parse_resource_name(resource_name: str, default_type: str) -> tuple[str, str]:
    if "." in resource_name:
        resref, ext = resource_name.rsplit(".", 1)
        return resref, ext.lower()
    return resource_name, default_type.lower()


def _resource_type_from_extension(extension: str, fallback: str) -> object:
    from pykotor.resource.type import ResourceType

    try:
        return ResourceType[extension.upper()]
    except KeyError:
        return ResourceType[fallback.upper()]


def _ensure_valid_tpc_bytes(data: bytes) -> bytes:
    from pykotor.resource.formats.tpc.tpc_auto import bytes_tpc
    from pykotor.resource.formats.tpc.tpc_data import TPC
    from pykotor.resource.type import ResourceType

    if data.startswith((b"TPC V1.0", b"TPC V1.1")):
        return data
    return bytes_tpc(TPC.from_blank(), ResourceType.TPC)


def _normalize_resource_bytes(restype: object, data: bytes) -> bytes:
    from pykotor.resource.type import ResourceType

    if restype == ResourceType.TPC:
        return _ensure_valid_tpc_bytes(data)
    return data


def _write_stream_resources(directory: Path, resources: Mapping[str, bytes] | None) -> None:
    if not resources:
        return
    directory.mkdir(parents=True, exist_ok=True)
    for resource_name, data in resources.items():
        resource_path = directory / resource_name
        resource_path.parent.mkdir(parents=True, exist_ok=True)
        resource_path.write_bytes(data)


def _write_archive_resources(archive: object, resources: Mapping[str, bytes], default_type: str) -> None:
    for resource_name, data in resources.items():
        resref, ext = _parse_resource_name(resource_name, default_type)
        restype = _resource_type_from_extension(ext, default_type)
        archive.set_data(resref, restype, _normalize_resource_bytes(restype, data))


def _write_talk_tables(
    root: Path,
    *,
    dialog_tlk_data: bytes | None,
    dialogf_tlk_data: bytes | None,
    dialog_entries: Sequence[str | tuple[str, str]] | None,
) -> None:
    from pykotor.common.language import Language
    from pykotor.resource.formats.tlk.tlk_auto import write_tlk
    from pykotor.resource.formats.tlk.tlk_data import TLK

    dialog_path = root / "dialog.tlk"
    dialogf_path = root / "dialogf.tlk"

    if dialog_tlk_data is not None:
        dialog_path.write_bytes(dialog_tlk_data)
        dialogf_path.write_bytes(dialog_tlk_data if dialogf_tlk_data is None else dialogf_tlk_data)
        return

    dialog_tlk = TLK(language=Language.ENGLISH)
    dialogf_tlk = TLK(language=Language.ENGLISH)

    entries = list(dialog_entries or ("", "", _DEFAULT_TLK_TEXT))
    for entry in entries:
        if isinstance(entry, tuple):
            text, sound = entry
        else:
            text, sound = entry, ""
        dialog_tlk.add(text, sound)
        dialogf_tlk.add(text, sound)

    write_tlk(dialog_tlk, dialog_path)
    write_tlk(dialogf_tlk, dialogf_path)


def _prepare_root_layout(root: Path, game: Game) -> None:
    from pykotor.common.misc import Game

    is_k2 = game in (Game.K2, Game.K2_XBOX, Game.K2_IOS, Game.K2_ANDROID)

    for subdir in ("data", "modules", "override", "streammusic", "streamsounds", "texturepacks", "lips"):
        (root / subdir).mkdir(parents=True, exist_ok=True)

    (root / ("streamvoice" if is_k2 else "streamwaves")).mkdir(parents=True, exist_ok=True)

    if is_k2:
        (root / "swkotor2.exe").touch()
        (root / "swkotor2.ini").touch()
        (root / "LocalVault").mkdir(exist_ok=True)
        (root / "LocalVault" / "test.bic").touch()
        (root / "LocalVault" / "testold.bic").touch()
    else:
        (root / "swkotor.exe").touch()
        (root / "swkotor.ini").touch()
        (root / "rims").mkdir(exist_ok=True)
        (root / "utils").mkdir(exist_ok=True)

    (root / "miles").mkdir(exist_ok=True)


def create_installation(
    path: Path,
    game: Game,
    *,
    with_override: bool = True,
    bif_resources: Mapping[str, Mapping[str, bytes]] | None = None,
    override_resources: Mapping[str, bytes] | None = None,
    modules_resources: Mapping[str, Mapping[str, bytes]] | None = None,
    voice_resources: Mapping[str, bytes] | None = None,
    music_resources: Mapping[str, bytes] | None = None,
    sound_resources: Mapping[str, bytes] | None = None,
    lips_resources: Mapping[str, Mapping[str, bytes]] | None = None,
    texture_tpa_resources: Mapping[str, bytes] | None = None,
    texture_tpb_resources: Mapping[str, bytes] | None = None,
    texture_tpc_resources: Mapping[str, bytes] | None = None,
    texture_gui_resources: Mapping[str, bytes] | None = None,
    patch_erf_resources: Mapping[str, bytes] | None = None,
    dialog_tlk_data: bytes | None = None,
    dialogf_tlk_data: bytes | None = None,
    dialog_entries: Sequence[str | tuple[str, str]] | None = None,
) -> Path:
    """Create a valid KotOR installation with real archives and TLK data."""
    from pykotor.common.misc import Game, ResRef
    from pykotor.resource.formats.bif.bif_auto import write_bif
    from pykotor.resource.formats.bif.bif_data import BIF
    from pykotor.resource.formats.erf.erf_auto import write_erf
    from pykotor.resource.formats.erf.erf_data import ERF, ERFType
    from pykotor.resource.formats.key.key_auto import write_key
    from pykotor.resource.formats.key.key_data import KEY
    from pykotor.resource.formats.rim.rim_auto import write_rim
    from pykotor.resource.formats.rim.rim_data import RIM
    from pykotor.resource.type import ResourceType

    root = Path(path)
    root.mkdir(parents=True, exist_ok=True)
    _prepare_root_layout(root, game)

    if not with_override:
        override_dir = root / "override"
        if override_dir.exists():
            for child in override_dir.iterdir():
                if child.is_file():
                    child.unlink()

    key = KEY()
    bif_resources = dict(bif_resources or {})
    is_k2 = game in (Game.K2, Game.K2_XBOX, Game.K2_IOS, Game.K2_ANDROID)

    if is_k2:
        bif_resources.setdefault("Dialogs.bif", {})
    else:
        bif_resources.setdefault("party.bif", {})
        bif_resources.setdefault("player.bif", {})

    for bif_filename, resources in bif_resources.items():
        bif_entry = key.add_bif(f"data/{bif_filename}")
        bif_index = len(key.bif_entries) - 1
        bif = BIF()

        for res_index, (resource_name, raw_data) in enumerate(resources.items()):
            resref, ext = _parse_resource_name(resource_name, "utc")
            restype = _resource_type_from_extension(ext, "utc")
            resource_id = key.calculate_resource_id(bif_index, res_index)
            bif.set_data(ResRef(resref), restype, _normalize_resource_bytes(restype, raw_data), res_id=resource_id)
            key.add_key_entry(ResRef(resref), restype, bif_index, res_index)

        bif_path = root / "data" / bif_filename
        write_bif(bif, bif_path)
        bif_entry.filesize = bif_path.stat().st_size

    write_key(key, root / "chitin.key")

    if with_override:
        _write_stream_resources(root / "override", override_resources)

    modules_dir = root / "modules"
    modules_resources = dict(modules_resources or {})
    if not is_k2:
        modules_resources.setdefault("global.mod", {})

    for module_filename, resources in modules_resources.items():
        module_path = modules_dir / module_filename
        lower_name = module_filename.lower()
        if lower_name.endswith(".rim"):
            archive = RIM()
            _write_archive_resources(archive, resources, "are")
            write_rim(archive, module_path)
        else:
            archive = ERF(erf_type=ERFType.MOD if lower_name.endswith(".mod") else ERFType.ERF)
            _write_archive_resources(archive, resources, "are")
            write_erf(archive, module_path, file_format=ResourceType.MOD if lower_name.endswith(".mod") else ResourceType.ERF)

    for mod_filename, resources in (lips_resources or {}).items():
        final_name = mod_filename if mod_filename.lower().endswith(".mod") else f"{mod_filename}.mod"
        archive = ERF(erf_type=ERFType.MOD)
        _write_archive_resources(archive, resources, "lip")
        write_erf(archive, root / "lips" / final_name, file_format=ResourceType.MOD)

    for filename, resources in (
        ("swpc_tex_tpa.erf", texture_tpa_resources),
        ("swpc_tex_tpb.erf", texture_tpb_resources),
        ("swpc_tex_tpc.erf", texture_tpc_resources),
        ("swpc_tex_gui.erf", texture_gui_resources),
    ):
        if not resources:
            continue
        archive = ERF()
        _write_archive_resources(archive, resources, "tpc")
        write_erf(archive, root / "texturepacks" / filename)

    if patch_erf_resources:
        archive = ERF(erf_type=ERFType.ERF)
        _write_archive_resources(archive, patch_erf_resources, "utc")
        write_erf(archive, root / "patch.erf", file_format=ResourceType.ERF)

    _write_stream_resources(root / ("streamvoice" if is_k2 else "streamwaves"), voice_resources)
    _write_stream_resources(root / "streammusic", music_resources)
    _write_stream_resources(root / "streamsounds", sound_resources)

    _write_talk_tables(
        root,
        dialog_tlk_data=dialog_tlk_data,
        dialogf_tlk_data=dialogf_tlk_data,
        dialog_entries=dialog_entries,
    )

    log.info("Created %s installation at: %s", "K2/TSL" if is_k2 else "K1", root)
    return root


def create_minimal_installation(
    path: Path,
    game: Game,
    *,
    validate_existing: bool = False,
) -> Path:
    """Scaffold a minimal but readable KotOR installation.

    If ``chitin.key`` already exists and ``validate_existing`` is False (default),
    returns immediately without creating or modifying other files. Use this when
    you only need a valid KEY and want to avoid overwriting an existing layout.

    When ``validate_existing=True`` (e.g. for tests or CI), a partial root is
    completed: directory layout and missing required files (e.g. dialog.tlk,
    dialogf.tlk) are created even if ``chitin.key`` is already present. Existing
    chitin.key and BIFs are not overwritten. Use this to ensure a minimal
    installation is fully usable when the root may have been partially created.
    """
    root = Path(path)
    root.mkdir(parents=True, exist_ok=True)
    if (root / "chitin.key").exists() and not validate_existing:
        return root
    if (root / "chitin.key").exists() and validate_existing:
        _prepare_root_layout(root, game)
        if not (root / "dialog.tlk").is_file():
            _write_talk_tables(root, dialog_tlk_data=None, dialogf_tlk_data=None, dialog_entries=("", "", _DEFAULT_TLK_TEXT))
        return root
    return create_installation(root, game)



from __future__ import annotations

import functools
import multiprocessing

from concurrent.futures import ProcessPoolExecutor
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Generic, Iterable, Iterator, TypeVar

from utility.system.app_process.task_consumer import TaskConsumer

if TYPE_CHECKING:
    from multiprocessing.synchronize import Event as MultiprocessingEvent


T = TypeVar("T")


class CustomProcessPoolExecutor(Generic[T]):
    """This is a class created for educational purposes only.

    Use concurrent.futures.ProcessPoolExecutor instead of this implementation.
    """

    _shared_state: ClassVar[dict[str, Any]] = {}

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        obj.__dict__ = cls._shared_state
        return obj

    def __init__(
        self,
        *,
        max_workers: int | None = multiprocessing.cpu_count() * 2,
        initializer: Callable[[], None] | None = None,
        initargs: tuple[Any, ...] = (),
    ):
        if hasattr(self, "singleton_initialized"):
            return
        self.max_workers: int = max_workers or multiprocessing.cpu_count() * 2
        self.task_queue: multiprocessing.JoinableQueue[Callable[..., Any]] = multiprocessing.JoinableQueue()
        self.result_queue: multiprocessing.Queue[T] = multiprocessing.Queue()
        self.workers: list[tuple[MultiprocessingEvent, TaskConsumer]] = []
        self._initializer: Callable[[], None] | None = initializer
        self.executor: ProcessPoolExecutor = ProcessPoolExecutor(
            max_workers=self.max_workers,
            initializer=self._initializer,
            initargs=initargs,
        )
        self._create_workers()
        self.singleton_initialized: bool = True

    def __del__(self):
        self.shutdown(wait=False)

    async def run(self):
        while True:
            task = self.task_queue.get(block=False, timeout=0.01)
            if task is None:
                break
            result = await self.executor.submit(task).result()
            self.result_queue.put(result)
            self.task_queue.task_done()

    def _create_workers(self):
        for i in range(self.max_workers):
            worker = TaskConsumer(
                task_queue=self.task_queue,
                result_queue=self.result_queue,
                name=f"TaskConsumer-{i}",
                daemon=True,
            )
            worker.start()
            stop_event = multiprocessing.Event()
            self.workers.append((stop_event, worker))

    def shutdown(
        self,
        *,
        wait: bool = True,
        timeout: float | None = None,
    ):
        for stop_event, worker in self.workers:
            stop_event.set()
            worker.join(timeout=timeout)
        if not wait:
            return
        for stop_event, _ in self.workers:
            stop_event.wait(timeout=timeout)

    def map(self, fn: Callable[..., T], *iterables: Iterable[Any]) -> Iterator[T]:
        return self.executor.map(fn, *iterables)


# Singleton instance
_instance: CustomProcessPoolExecutor | None = None


def get_instance(*args, **kwargs) -> CustomProcessPoolExecutor:
    global _instance  # noqa: PLW0603
    if _instance is None:
        _instance = CustomProcessPoolExecutor(*args, **kwargs)
    return _instance


if __name__ == "__main__":

    def example_task(x: Any) -> Any:
        return x * x

    executor1 = get_instance(max_workers=1)
    executor2 = get_instance(max_workers=2)  # This will return the same instance as executor1

    results = list(
        executor1.map(
            functools.partial(print, "Hello World"),
            [(x, i) for i in range(10) for x in range(10)],
        )
    )

    print("Results:", results)
    print("executor1 is executor2:", executor1 is executor2)  # This should print True



"""Panda3D implementation of PyKotor engine interfaces.

This module contains Panda3D-specific implementations of the abstract engine interfaces
defined in pykotor.engine.

References:
----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        Libraries/PyKotor/src/pykotor/engine - Abstract interfaces


"""

from __future__ import annotations

from pykotor.engine.panda3d.engine import KotorEngine, create_engine
from pykotor.engine.panda3d.scene_graph import Panda3DSceneGraph
from pykotor.engine.panda3d.animation import (
    Panda3DAnimationController,
    Panda3DPositionController,
    Panda3DOrientationController,
    Panda3DScaleController,
    Panda3DColorController,
    Panda3DAlphaController,
    Panda3DAnimationState,
    Panda3DAnimationManager,
    create_panda3d_controller,
)
from pykotor.engine.panda3d.mdl_loader import MDLLoader
from pykotor.engine.panda3d.materials import Panda3DMaterial, Panda3DMaterialManager

__all__ = [
    "KotorEngine",
    "create_engine",
    "Panda3DSceneGraph",
    "Panda3DAnimationController",
    "Panda3DPositionController",
    "Panda3DOrientationController",
    "Panda3DScaleController",
    "Panda3DColorController",
    "Panda3DAlphaController",
    "Panda3DAnimationState",
    "Panda3DAnimationManager",
    "create_panda3d_controller",
    "MDLLoader",
    "Panda3DMaterial",
    "Panda3DMaterialManager",
]



from __future__ import annotations

import inspect
import os
import sys
import traceback
import types

from contextlib import suppress
from contextvars import ContextVar
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar, cast

if TYPE_CHECKING:
    from collections.abc import Callable

    from typing_extensions import Literal


def universal_simplify_exception(
    e: BaseException,
) -> tuple[str, str]:
    """Simplify exceptions into a standardized format.

    Args:
    ----
        e: Exception - The exception to simplify

    Returns:
    -------
        error_name: str - The name of the exception
        error_message: str - A human-readable message for the exception

    Processing Logic:
    ----------------
        - Extract the exception name from the type
        - Handle specific exception types differently
        - Try common exception attributes for a message
        - Return a general fallback if nothing else better was determined.
    """
    error_name: str = e.__class__.__name__

    # Handle FileNotFoundError, which has 'filename' attribute
    if isinstance(e, FileNotFoundError):
        filename = getattr(e, "filename", getattr(e, "filename2", None))
        if filename:
            return error_name, f"Could not find the file: '{filename}'"
        if len(e.args) == 1:
            return error_name, f"Could not find the file: '{e.args[0]}'"
        if len(e.args) > 1:
            return error_name, f"{e.args[1]}: {e.args[0]}"

    # Handle PermissionError, which may have a 'filename' attribute
    if isinstance(e, PermissionError):
        return error_name, f"Permission Denied: {e.filename if hasattr(e, 'filename') else e.args[0] if e.args else ''}"

    # Handle TimeoutError
    if isinstance(e, TimeoutError):
        return error_name, f"Operation timed out: {e.args[0] if e.args else ''}"

    # Handle InterruptedError, which may have an 'errno' attribute
    if isinstance(e, InterruptedError):
        return error_name, f"Interrupted: {e.errno}"

    # Handle ConnectionError, which may have a 'request' attribute if it's from the `requests` library
    if isinstance(e, ConnectionError):
        return error_name, f"Connection Error: {getattr(e, 'request', lambda: {'method': e.args[0]})().get('method', '')}"

    # Add more oddball exception handling here as needed

    error_messages = [f"- {attr}: {getattr(e, attr)}" for attr in ["strerror", "filename", "filename1", "filename2", "message", "reason"] if getattr(e, attr, None)]
    if error_messages:
        error_details = "\n".join(error_messages)
        return error_name, f"{e}:\n\nDetails:\n{error_details}"

    err_str = str(e)
    args = getattr(e, "args", [])
    if args and isinstance(args, list):
        err_msg_index: int | Literal[False] = err_str in args and args.index(err_str)
        if err_msg_index:
            del args[err_msg_index]
        if args:
            err_str += "\n  Information:"
            for arg in args:
                err_str += f"\n    {arg}"
    return error_name, err_str


# Get default module attributes to filter out built-ins
DEFAULT_ATTRS: set[str] = set(dir(sys.modules["builtins"]))

IGNORE_ATTRS: set[str] = {
    "__file__",
    "__cached__",
    "__builtins__",
}


class CustomAssertionError(AssertionError):
    def __init__(self, *args, **kwargs):
        self.message: str = kwargs.get("message", args[0])
        super().__init__(*args, **kwargs)


def is_builtin_class_instance(obj: Any) -> bool:
    """Check if the object is an instance of a built-in class."""
    return obj.__class__.__module__ in ("builtins", "__builtin__")


_CURRENTLY_PROCESSING: ContextVar[list] = ContextVar("_currently_processing", default=[])


def safe_repr(
    obj: Any,
    max_length: int = 200,
    indent_level: int = 0,
    max_depth: int = 3,
    _depth: int = 0,
    *,
    ignore_builtins: bool = True,
) -> str:
    """Safely generate a repr string for objects without a custom __repr__, with line wrapping and indentation."""
    if ignore_builtins and is_builtin_class_instance(obj):
        try:
            obj_repr = repr(obj)
            # Truncate if necessary
            if len(obj_repr) > max_length:
                return f"{obj_repr[:max_length]}..."
        except Exception:  # noqa: BLE001
            return object.__repr__(obj)
        else:
            return obj_repr
    indent: str = "    "  # Define the indentation unit (4 spaces).

    # Retrieve the stack of objects currently being processed
    current_stack = _CURRENTLY_PROCESSING.get()
    obj_id = id(obj)

    # Check for recursion - if this object is already in the stack
    for frame in current_stack:
        if obj_id == frame["id"]:
            # Finish the partial representation for this object and return
            base_indent = indent * frame["indent_level"]
            return f"{frame['representation']}...\n{base_indent})"

    if _depth > max_depth:
        try:
            obj_repr = repr(obj)
            # Truncate if necessary
            if len(obj_repr) > max_length:
                return f"{obj_repr[:max_length]}..."
        except Exception:  # noqa: BLE001
            return object.__repr__(obj)
        else:
            return obj_repr

    try:
        # Initialize the representation for this object and add it to the stack
        base_indent = indent * indent_level
        next_indent = indent * (indent_level + 1)
        representation: str = f"{obj.__class__.__name__}(\n{next_indent}"
        current_stack.append({"id": obj_id, "indent_level": indent_level, "representation": representation})
        _CURRENTLY_PROCESSING.set(current_stack)

        if hasattr(obj, "__class__") and obj.__class__.__repr__ is not object.__repr__:
            # Call the object's __repr__ with _is_safe_repr_call set to True
            try:
                return repr(obj)
            except Exception:  # noqa: BLE001
                # Fallback to the base object __repr__ if an error occurs
                representation += object.__repr__(obj)
            return representation

        attrs: list[str] = []
        for attr_name in cast("dict[str, Any]", vars(obj)):
            attr_value = getattr(obj, attr_name)
            if not attr_name.startswith("__") and not callable(attr_value):
                try:
                    this_repr = safe_repr(attr_value, max_length, indent_level + 1, _depth=_depth + 1)
                    # Concatenate attribute name and its representation with appropriate indentation
                    attr_repr = f"{attr_name}={this_repr}"
                    # Check if current attribute representation exceeds the max length
                    if len(attr_repr) > max_length:
                        attr_repr = f"{attr_repr[:max_length]}..."
                    attrs.append(attr_repr)
                except Exception:  # noqa: BLE001
                    attrs.append(f"{attr_name}={object.__repr__(attr_value)}")
        joined_attrs = (",\n" + next_indent).join(attrs)
        final_repr = f"{representation}{joined_attrs}\n{base_indent})" if attrs else f"{representation}{base_indent})"
    except Exception:  # noqa: BLE001
        return object.__repr__(obj)
    else:
        return final_repr
    finally:
        # Always remove the object from the stack to avoid leaks
        current_stack.pop()
        _CURRENTLY_PROCESSING.set(current_stack)


def format_var_str(
    var: str,
    val: Any,
    max_length: int = 512,
) -> str | None:
    """Format variable and its value into a string, handling exceptions and length."""
    if var in DEFAULT_ATTRS or var in IGNORE_ATTRS:
        return None

    val_repr: str | object
    val_str: str | object
    exc = None
    unique_sentinel = object()
    try:
        val_str = str(val)
        if len(val_str) > max_length:
            val_str = f"{val_str[:max_length]}...<truncated>"
    except Exception as e:  # pylint: disable=W0718  # noqa: BLE001
        val_str = unique_sentinel
        exc = e

    try:
        val_repr = safe_repr(val)
        if len(val_repr) > max_length * 2:
            val_repr = f"{val_repr[:max_length]}...<truncated>"
    except Exception:  # pylint: disable=W0718  # noqa: BLE001
        val_repr = unique_sentinel

    display_value: str | object = val_repr
    if display_value is unique_sentinel:
        display_value = val_str
    if display_value is unique_sentinel:
        display_value = f"<Error in repr({var}): {exc}>"

    return f"  {var} = {display_value}"


def format_frame_info(
    frame_info: inspect.FrameInfo,
) -> list[str]:
    """Extract and format information from a frame."""
    (
        frame,
        filename,
        line_no,
        function,
        code_context,
        index,
    ) = frame_info
    code_context = ""  # type: ignore[assignment]
    # code_context = f"\nContext [{', '.join(code_context)}] " if code_context else ""  # type: ignore[assignment]
    detailed_message: list[str] = [f"\nFunction '{function}' at {filename}:{line_no}:{code_context}"]
    for var, val in frame.f_locals.items():
        formatted_var: str | None = format_var_str(var, val)
        if formatted_var:
            detailed_message.append(formatted_var)
    return detailed_message


@lru_cache(maxsize=128)
def format_exception_with_variables(
    exc: BaseException,
    etype: type[BaseException] | None = None,
    tb: types.TracebackType | None = None,
    message: str = "",
) -> str:
    etype = type(exc) if etype is None else etype
    tb = exc.__traceback__ if tb is None else tb

    # Check if the arguments are of the correct type
    if not issubclass(etype, BaseException):
        msg = f"{etype!r} is not an exception class"
        raise TypeError(msg)
    if not isinstance(exc, BaseException):
        msg = f"{exc!r} is not an exception instance"
        raise TypeError(msg)
    if not isinstance(tb, types.TracebackType):
        with suppress(Exception):
            # Get the current stack frames
            current_stack = inspect.stack()
            if current_stack:
                # Reverse the stack to have the order from caller to callee
                current_stack = current_stack[1:][::-1]
                fake_traceback = None
                for frame_info in current_stack:
                    frame = frame_info.frame
                    fake_traceback = types.TracebackType(fake_traceback, frame, frame.f_lasti, frame.f_lineno)
                exc = exc.with_traceback(fake_traceback)
                # Now exc has a traceback :)
                tb = exc.__traceback__

    # Construct the stack trace using traceback
    formatted_traceback: str = "".join(traceback.format_exception(etype, exc, tb))

    # Construct a detailed message with variables from all stack frames
    detailed_message: list[str] = [
        formatted_traceback,
        f"{etype.__name__}: {exc}",
        "Stack Trace Variables:",
    ]
    if tb is None:
        detailed_message.append("<No stack information available>")
        return "\n".join(detailed_message)

    # Capture the current stack trace
    frames: list[inspect.FrameInfo] = inspect.getinnerframes(tb, context=5)
    for frame_info in frames:
        detailed_message.extend(format_frame_info(frame_info))
    if exc.__cause__ is not None:
        detailed_message.append("This is the original exception:")
        detailed_message.extend(
            format_exception_with_variables(
                exc.__cause__,
                message="Causing Exception's Stack Trace Variables:",
            ).split("\n")
        )

    detailed_message.append(formatted_traceback)
    return "\n".join(detailed_message)


def is_assertion_removal_enabled() -> bool:
    return sys.flags.optimize >= 1


IT = TypeVar("IT")


def enforce_instance_cast(
    obj: object,
    type_: type[IT],
) -> IT:
    instance_check: bool = isinstance(obj, type_)
    if is_assertion_removal_enabled():
        # don't enforce the instance check if the assertion optimizers are being used.
        if not instance_check:
            print(format_exception_with_variables(AssertionError("isinstance(obj, type_)")))  # noqa: T201
        return obj  # type: ignore[return-value]

    assert_with_variable_trace(isinstance(obj, type_), "enforce_is_instance failed.")
    return obj  # type: ignore[return-value]


def assert_with_variable_trace(
    condition: bool,  # noqa: FBT001
    message: str = "Assertion Failed",
):
    if condition:
        return
    # Capture the current stack trace
    frames: list[inspect.FrameInfo] = inspect.getouterframes(inspect.currentframe())

    # Get the line of code calling assert_with_variable_trace
    calling_frame_record: inspect.FrameInfo = frames[1]
    (
        frame,
        filename,
        line_no,
        function,
        code_context,
        frame_type,
    ) = calling_frame_record
    line_of_code: str = code_context[0].strip() if code_context else "Unknown condition"

    # Construct a detailed message with variables from all stack frames
    detailed_message: list[str] = [
        f"{message}: Expected condition '{line_of_code}' failed at {filename}:{line_no}.",
        "Stack Trace Variables:",
    ]
    for frame_info in frames:
        detailed_message.extend(format_frame_info(frame_info))

    full_message: str = "\n".join(detailed_message)

    # Raise an exception with the detailed message
    raise AssertionError(full_message)


RT = TypeVar("RT")
unique_sentinel = object()


def with_variable_trace(
    exception_types: type[Exception] | tuple[type[Exception], ...] = Exception,
    return_type: type[RT] = unique_sentinel,  # pyright: ignore[reportGeneralTypeIssues, assignment]  # type: ignore[assignment]
    *,
    action: Literal["print", "stderr"] = "print",
    log: bool = True,
    rethrow: bool = False,
) -> Callable[[Callable[..., RT]], Callable[..., RT | None]]:
    # Set default to Exception if no specific types are provided
    if not exception_types:
        exception_types = (Exception,)
    if not isinstance(exception_types, tuple):
        exception_types = (exception_types,)
    exception_types = (*exception_types, CustomAssertionError)

    def decorator(f: Callable[..., RT]) -> Callable[..., RT | None]:
        def wrapper(*args, **kwargs) -> RT | None:
            try:
                result: RT = f(*args, **kwargs)
                if return_type is not unique_sentinel and not isinstance(result, return_type):
                    msg = f"Return type of '{f.__name__}' must be {return_type.__name__}, got {result.__class__}: {result!r}: {result}"
                    raise CustomAssertionError(msg)
            except exception_types as e:
                detailed_message: list[str] = [
                    f"Exception caught in function '{f.__name__}': {e}",
                    "Stack Trace Variables:",
                ]
                for frame_info in inspect.getouterframes(inspect.currentframe()):
                    # if frame_info.function != f.__name__:  # Why did I add this?
                    #    continue
                    detailed_message.extend(format_frame_info(frame_info))
                if e.__cause__ is not None:
                    detailed_message.append("This is the original exception:")
                    detailed_message.extend(format_exception_with_variables(e.__cause__, message="Causing Exception's Stack Trace Variables:").split("\n"))

                full_message: str = "\n".join(detailed_message)

                if action == "stderr":
                    if sys.stderr is None:
                        sys.stderr = open(os.devnull, "w")  # noqa: PTH123, SIM115, PLW1514  # pylint: disable=all
                    print(full_message, sys.stderr)  # noqa: T201
                elif action == "print":
                    print(full_message)  # noqa: T201
                if log:
                    with Path("errorlog.txt", encoding="utf-8").open("a", encoding="utf-8") as outfile:
                        outfile.write(full_message)
                if rethrow:
                    # Raise an exception with the detailed message
                    if isinstance(e, CustomAssertionError):
                        raise AssertionError(e.message) from None
                    raise e.__class__(full_message) from e

                return None
            else:
                return result

        return wrapper

    return decorator



"""Init command implementation."""

from __future__ import annotations

import subprocess

from argparse import Namespace
from pathlib import Path
from typing import TYPE_CHECKING

from utility.misc import ensure_directory_exists

if TYPE_CHECKING:
    from logging import Logger


DEFAULT_CONFIG_TEMPLATE = """# Pykotorcli Package Configuration
# This file uses TOML format and is compatible with cli syntax

[package]
name = "{package_name}"
description = "{description}"
version = "1.0.0"
author = "{author}"

# Default file pattern - inherited by targets if not specified
file = "$target.mod"

  [package.sources]
  include = "src/**/*.{{nss,json,ncs,utc,uti,utp,utd,ute,utm,uts,utt,utw,git,are,ifo,dlg,gui}}"
  exclude = "**/test_*.nss"

  [package.rules]
  "*.nss" = "src/scripts"
  "*.ncs" = "src/scripts"
  "*.utc" = "src/blueprints/creatures"
  "*.uti" = "src/blueprints/items"
  "*.utp" = "src/blueprints/placeables"
  "*.utd" = "src/blueprints/doors"
  "*.ute" = "src/blueprints/encounters"
  "*.utm" = "src/blueprints/merchants"
  "*.uts" = "src/blueprints/sounds"
  "*.utt" = "src/blueprints/triggers"
  "*.utw" = "src/blueprints/waypoints"
  "*.dlg" = "src/dialogs"
  "*.git" = "src/areas"
  "*.are" = "src/areas"
  "*.ifo" = "src"
  "*.gui" = "src/gui"
  "*" = "src"

[target]
name = "default"
file = "{target_file}"
description = "Default target"
"""


def prompt(message: str, default: str = "") -> str:
    """Prompt user for input with a default value."""
    if default:
        message = f"{message} [{default}]"
    message += ": "

    response = input(message).strip()
    return response if response else default


def _gather_package_info(args: Namespace, target_dir: Path, logger: Logger) -> tuple[str, str, str, str]:
    """Gather package information from user or use defaults."""
    if args.default:
        package_name = target_dir.name
        description = ""
        author = ""
        target_file = f"{package_name}.mod"
    else:
        logger.info("\nPackage Configuration")
        logger.info("---------------------")
        package_name = prompt("Package name", target_dir.name)
        description = prompt("Description", "")
        author = prompt("Author", "")

        logger.info("\nTarget Configuration")
        logger.info("--------------------")
        target_file = prompt("Target filename (e.g., mymod.mod, myhak.hak)", f"{package_name}.mod")

    return package_name, description, author, target_file


def _create_directory_structure(target_dir: Path, logger: Logger) -> None:
    """Create the standard directory structure for a pykotorcli package."""
    src_dir = target_dir / "src"
    src_dir.mkdir(exist_ok=True)
    logger.info(f"Created {src_dir}")  # noqa: G004

    subdirs = [
        "src/scripts",
        "src/dialogs",
        "src/areas",
        "src/blueprints/creatures",
        "src/blueprints/items",
        "src/blueprints/placeables",
        "src/blueprints/doors",
        "src/blueprints/encounters",
        "src/blueprints/merchants",
        "src/blueprints/sounds",
        "src/blueprints/triggers",
        "src/blueprints/waypoints",
        "src/gui",
    ]

    for subdir in subdirs:
        ensure_directory_exists(target_dir / subdir)

    kotorcli_dir = target_dir / ".pykotorcli"
    kotorcli_dir.mkdir(exist_ok=True)


def _create_gitignore(target_dir: Path, logger: Logger) -> None:
    """Create .gitignore file if it doesn't exist."""
    gitignore_content = """# Pykotorcli
.pykotorcli/cache/
.pykotorcli/user.cfg
*.log

# Build output
*.mod
*.erf
*.hak
*.rim
*.bif
*.key

# Python
__pycache__/
*.py[cod]
*$py.class
*.so
.Python

# Compiled scripts
*.ncs
"""
    gitignore_path = target_dir / ".gitignore"
    if not gitignore_path.exists():
        gitignore_path.write_text(gitignore_content, encoding="utf-8")
        logger.info(f"Created {gitignore_path}")  # noqa: G004


def _initialize_git_repo(target_dir: Path, logger: Logger) -> None:
    """Initialize git repository if requested."""
    try:
        subprocess.run(
            ["git", "init"],
            cwd=target_dir,
            check=True,
            capture_output=True,
        )
        logger.info("Initialized git repository")

        gitattributes_content = """# Auto detect text files and perform LF normalization
* text=auto

# Scripts should use LF
*.nss text eol=lf
*.ncs binary
"""
        gitattributes_path = target_dir / ".gitattributes"
        gitattributes_path.write_text(gitattributes_content, encoding="utf-8")
        logger.info(f"Created {gitattributes_path}")  # noqa: G004

    except subprocess.CalledProcessError:
        logger.warning("Failed to initialize git repository (git not found or error occurred)")
    except FileNotFoundError:
        logger.warning("Git not found, skipping repository initialization")


def _unpack_initial_file(args: Namespace, init_file: str, logger: Logger) -> None:
    """Unpack initial file if specified."""
    logger.info(f"Unpacking initial file: {init_file}")  # noqa: G004
    from pykotor.cli.commands.unpack import cmd_unpack  # noqa: PLC0415

    unpack_args = Namespace(
        target=None,
        file=init_file,
        unpack_file=init_file,
        removeDeleted=False,
        debug=args.debug if hasattr(args, "debug") else False,
        quiet=args.quiet if hasattr(args, "quiet") else False,
        verbose=args.verbose if hasattr(args, "verbose") else False,
        no_color=args.no_color if hasattr(args, "no_color") else False,
    )
    cmd_unpack(unpack_args, logger)


def cmd_init(args: Namespace, logger: Logger) -> int:
    """Handle init command - create a new pykotorcli package.

    Args:
    ----
        args: Parsed command line arguments
        logger: Logger instance

    Returns:
    -------
        Exit code (0 for success, non-zero for error)
    """
    target_dir = Path(args.dir if args.dir else ".").resolve()
    ensure_directory_exists(target_dir)

    config_path = target_dir / "pykotorcli.cfg"
    if config_path.exists():
        logger.warning(f"Package already initialized at {target_dir}")  # noqa: G004
        overwrite = prompt("Overwrite existing configuration? (yes/no)", "no")
        if overwrite.lower() not in ("yes", "y"):
            logger.info("Initialization cancelled")
            return 0

    logger.info(f"Initializing pykotorcli package in {target_dir}")  # noqa: G004

    package_name, description, author, target_file = _gather_package_info(args, target_dir, logger)

    config_content = DEFAULT_CONFIG_TEMPLATE.format(
        package_name=package_name,
        description=description,
        author=author,
        target_file=target_file,
    )
    config_path.write_text(config_content, encoding="utf-8")
    logger.info(f"Created {config_path}")  # noqa: G004

    _create_directory_structure(target_dir, logger)
    _create_gitignore(target_dir, logger)

    if args.vcs == "git":
        _initialize_git_repo(target_dir, logger)

    init_file = args.init_file or args.file
    if init_file:
        _unpack_initial_file(args, init_file, logger)

    logger.info(f"\nPackage initialized successfully in {target_dir}")  # noqa: G004
    logger.info("Next steps:")
    logger.info("  1. Edit pykotorcli.cfg to configure your package")
    logger.info("  2. Place your source files in the src/ directory")
    logger.info("  3. Run 'pykotorcli list' to see available targets")
    logger.info("  4. Run 'pykotorcli pack' to build your module")

    return 0



"""2DA modification algorithms for TSLPatcher/HoloPatcher.

This module implements 2DA modification logic for applying patches from changes.ini files.
Handles row/column additions, cell modifications, and memory token resolution.

References:
----------
        Based on swkotor.exe 2DA structure:
        - C2DA::Load2DArray @ 0x004143b0 - Loads 2DA file from resource
        - C2DA::Unload2DArray @ 0x004139e0 - Unloads 2DA data
        Original BioWare engine binaries
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/TSLPatcher/tree/master/TSLPatcher.pl (unfinished perl rewrite of TSLPatcher)
        https://github.com/th3w1zard1/HoloPatcher.NET/tree/master/src/TSLPatcher.Core/Mods/2DA/
        https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET.Patcher/


"""

from __future__ import annotations

from abc import ABC, abstractmethod
from enum import IntEnum
from pathlib import PureWindowsPath
from typing import TYPE_CHECKING, Mapping

from loggerplus import RobustLogger
from pykotor.resource.formats.twoda import bytes_2da, read_2da
from pykotor.tslpatcher.mods.template import PatcherModifications

if TYPE_CHECKING:
    from typing_extensions import Literal  # pyright: ignore[reportMissingModuleSource]

    from pykotor.common.misc import Game
    from pykotor.resource.formats.twoda import TwoDA, TwoDARow
    from pykotor.resource.type import SOURCE_TYPES
    from pykotor.tslpatcher.logger import PatchLogger
    from pykotor.tslpatcher.memory import PatcherMemory


class CriticalError(Exception): ...


class WarningError(Exception): ...


class TargetType(IntEnum):
    ROW_INDEX = 0
    ROW_LABEL = 1
    LABEL_COLUMN = 2


class Target:
    def __init__(
        self,
        target_type: TargetType,
        value: str | int | RowValue2DAMemory | RowValueTLKMemory,
    ):
        self.target_type: TargetType = target_type
        self.value: str | int | RowValueTLKMemory | RowValue2DAMemory = value

        if target_type == TargetType.ROW_INDEX and isinstance(value, str):
            msg = "Target value must be int if type is row index."
            raise ValueError(msg)

    def __str__(self):
        value_str = str(self.value)
        class_name = self.value.__class__.__name__
        if class_name in ("RowValue2DAMemory", "RowValueTLKMemory"):
            value_str = repr(self.value)
        type_names = {
            TargetType.ROW_INDEX: "row index",
            TargetType.ROW_LABEL: "row label",
            TargetType.LABEL_COLUMN: "label column",
        }
        type_name = type_names.get(self.target_type, self.target_type.name)
        return f"Target({type_name}={value_str})"

    def __repr__(self):
        return f"{self.__class__.__name__}(target_type={self.target_type.__class__.__name__}.{self.target_type.name}, value={self.value!r})"

    def search(
        self,
        twoda: TwoDA,
        memory: PatcherMemory,
    ) -> TwoDARow | None:
        """Searches a TwoDA for a row matching the target.

        Args:
        ----
            twoda: TwoDA - The TwoDA to search
            memory: PatcherMemory - The memory context

        Returns:
        -------
            TwoDARow | None - The matching row if found, else None

        Processing Logic:
        ----------------
            - Checks target_type and searches twoda accordingly
            - For row index, gets row directly
            - For row label, finds row by label
            - For label column, checks for label column, then iterates rows to find match
            - Returns matching row or None.
        """
        value: str | int = self.value.value(memory, twoda, None) if isinstance(self.value, (RowValueTLKMemory, RowValue2DAMemory)) else self.value
        source_row: TwoDARow | None = None
        if self.target_type == TargetType.ROW_INDEX:
            source_row = twoda.get_row(int(value))
        elif self.target_type == TargetType.ROW_LABEL:
            source_row = twoda.find_row(str(value))
        elif self.target_type == TargetType.LABEL_COLUMN:
            if "label" not in twoda.get_headers():
                msg = f"'label' could not be found in the twoda's headers: ({self.target_type.name}, {value})"
                raise WarningError(msg)
            if value not in twoda.get_column("label"):
                msg = f"The value '{value}' could not be found in the twoda's columns"
                raise WarningError(msg)
            for row in twoda:
                if row.get_string("label") == value:
                    source_row = row

        return source_row


# region Value Returners
class RowValue(ABC):
    def __str__(self):
        return f"{self.__class__.__name__}()"

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    @abstractmethod
    def value(self, memory: PatcherMemory, twoda: TwoDA, row: TwoDARow | None) -> str:
        """Returns the value of the row.

        Args:
        ----
            memory: PatcherMemory - The memory context
            twoda: TwoDA - The TwoDA to search
            row: TwoDARow | None - The row to search
        """


class RowValueConstant(RowValue):
    def __init__(self, string: str):
        self.string: str = string

    def __str__(self):
        return f"Constant('{self.string}')"

    def __repr__(self):
        return f"{self.__class__.__name__}(string='{self.string}')"

    def value(self, memory: PatcherMemory, twoda: TwoDA, row: TwoDARow | None) -> str:
        return self.string


class RowValue2DAMemory(RowValue):
    def __init__(self, token_id: int):
        self.token_id: int = token_id

    def __str__(self):
        return f"2DAMEMORY{self.token_id}"

    def __repr__(self):
        return f"{self.__class__.__name__}(token_id={self.token_id})"

    def value(self, memory: PatcherMemory, twoda: TwoDA, row: TwoDARow | None) -> str:
        memory_val: str | PureWindowsPath | None = memory.memory_2da.get(self.token_id)
        if memory_val is None:
            msg = f"2DAMEMORY{self.token_id} was not defined before use."
            raise KeyError(msg)
        if isinstance(memory_val, PureWindowsPath):
            msg = f"!FieldPath cannot be used in 2DAList patches, got '{memory_val}'"
            raise TypeError(msg)
        return memory_val


class RowValueTLKMemory(RowValue):
    def __init__(self, token_id: int):
        self.token_id: int = token_id

    def __str__(self):
        return f"StrRef{self.token_id}"

    def __repr__(self):
        return f"{self.__class__.__name__}(token_id={self.token_id})"

    def value(self, memory: PatcherMemory, twoda: TwoDA, row: TwoDARow | None) -> str:
        memory_val: int | None = memory.memory_str.get(self.token_id)
        if memory_val is None:
            msg = f"StrRef{self.token_id} was not defined before use."
            raise KeyError(msg)
        return str(memory_val)


class RowValueHigh(RowValue):
    """
    Attributes:
    ----------
    column: Column to get the max integer from. If None it takes it from the Row Label.
    """  # noqa: D212, D415

    def __init__(self, column: str | None):
        self.column: str | None = column

    def __str__(self):
        if self.column is None:
            return "High(label_max)"
        return f"High(column_max('{self.column}'))"

    def __repr__(self):
        return f"{self.__class__.__name__}(column='{self.column}')"

    def value(self, memory: PatcherMemory, twoda: TwoDA, row: TwoDARow | None) -> str:
        """Returns the maximum value in a column or overall label.

        Args:
        ----
            memory: PatcherMemory object
            twoda: TwoDA object
            row: TwoDARow object or None

        Returns:
        -------
            str: String representation of maximum value

        Processing Logic:
        ----------------
            - If column is not None, return maximum value in that column
            - Else return overall maximum label value.
        """
        return str(twoda.label_max()) if self.column is None else str(twoda.column_max(self.column))


class RowValueRowIndex(RowValue):
    """The row index of the row."""

    def __init__(self): ...

    def __str__(self):
        return "RowIndex()"

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def value(self, memory: PatcherMemory, twoda: TwoDA, row: TwoDARow | None) -> str:
        return "" if row is None else str(twoda.row_index(row))


class RowValueRowLabel(RowValue):
    """The row label of the row."""

    def __init__(self): ...

    def __str__(self):
        return "RowLabel()"

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def value(self, memory: PatcherMemory, twoda: TwoDA, row: TwoDARow | None) -> str:
        return "" if row is None else row.label()


class RowValueRowCell(RowValue):
    """The value of a cell in the row."""

    def __init__(self, column: str):
        self.column: str = column

    def __str__(self):
        return f"RowCell('{self.column}')"

    def __repr__(self):
        return f"{self.__class__.__name__}(column='{self.column}')"

    def value(self, memory: PatcherMemory, twoda: TwoDA, row: TwoDARow | None) -> str:
        """Returns the value of a cell in the row."""
        return "" if row is None else row.get_string(self.column)


# endregion


# region Modify 2DA
class Modify2DA(ABC):
    def __str__(self):
        return f"{self.__class__.__name__}()"

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    @abstractmethod
    def __init__(self): ...

    def _unpack(
        self,
        cells: dict[str, RowValue],
        memory: PatcherMemory,
        twoda: TwoDA,
        row: TwoDARow,
    ) -> dict[str, str]:
        return {column: value.value(memory, twoda, row) for column, value in cells.items()}

    @abstractmethod
    def apply(
        self,
        twoda: TwoDA,
        memory: PatcherMemory,
    ): ...


class ChangeRow2DA(Modify2DA):
    """Changes an existing row.

    Target row can either be the Row Index, Row Label, or value under the "label" column where applicable.

    Attributes:
    ----------
        target: The row to change.
        modifiers: For the row, sets a cell under column KEY to have the text VALUE.
    """

    def __init__(
        self,
        identifier: str,
        target: Target,
        cells: Mapping[str, RowValue],
        store_2da: dict[int, RowValue] | None = None,
        store_tlk: dict[int, RowValue] | None = None,
    ):
        self.identifier: str = identifier
        self.target: Target = target
        self.cells: dict[str, RowValue] = dict(cells)
        self.store_2da: dict[int, RowValue] = {} if store_2da is None else store_2da
        self.store_tlk: dict[int, RowValue] = {} if store_tlk is None else store_tlk

        self._row: TwoDARow | None = None

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(identifier={self.identifier!r}, "
            f"target={self.target!r}, cells={self.cells!r}, "
            f"store_2da={self.store_2da!r}, store_tlk={self.store_tlk!r})"
        )

    def apply(self, twoda: TwoDA, memory: PatcherMemory):
        """Applies a ChangeRow patch to a TwoDA.

        Args:
        ----
            twoda: TwoDA - The TwoDA to apply the patch to
            memory: PatcherMemory - The memory context

        Processing Logic:
        ----------------
            - Searches for the target row
            - Unpacks the cells and applies them to the target row
            - Stores any 2DA or TLK values in the memory context.
        """
        source_row: TwoDARow | None = self.target.search(twoda, memory)

        if source_row is None:
            msg = f"The source row was not found during the search: ({self.target.target_type.name}, {self.target.value})"
            raise WarningError(msg)

        cells: dict[str, str] = self._unpack(self.cells, memory, twoda, source_row)
        source_row.update_values(cells)

        for token_id, value in self.store_2da.items():
            memory.memory_2da[token_id] = value.value(memory, twoda, source_row)

        for token_id, value in self.store_tlk.items():
            memory.memory_str[token_id] = int(value.value(memory, twoda, source_row))


class AddRow2DA(Modify2DA):
    """Adds a new row.

    Attributes:
    ----------
        modifiers: For the row, sets a cell under column KEY to have the text VALUE.
    """

    def __init__(  # noqa: PLR0913
        self,
        identifier: str,
        exclusive_column: str | None,
        row_label: str | None,
        cells: Mapping[str, RowValue],
        store_2da: dict[int, RowValue] | None = None,
        store_tlk: dict[int, RowValue] | None = None,
    ):
        self.identifier: str = identifier
        self.exclusive_column: str | None = exclusive_column
        self.row_label: str | None = row_label
        self.cells: dict[str, RowValue] = dict(cells)
        self.store_2da: dict[int, RowValue] = {} if store_2da is None else store_2da
        self.store_tlk: dict[int, RowValue] = {} if store_tlk is None else store_tlk

        self._row: TwoDARow | None = None

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(identifier={self.identifier!r}, "
            f"exclusive_column={self.exclusive_column!r}, row_label={self.row_label!r}, "
            f"cells={self.cells!r}, store_2da={self.store_2da!r}, "
            f"store_tlk={self.store_tlk!r})"
        )

    def apply(self, twoda: TwoDA, memory: PatcherMemory):
        """Applies an AddRow patch to a TwoDA.

        Args:
        ----
            twoda: TwoDA - The Two Dimensional Array to apply the patch to.
            memory: PatcherMemory - The memory context.

        Processing Logic:
        ----------------
            - Finds the target row to apply the patch to based on an optional exclusive column
            - If no target row is found, a new row is added
            - The cells are unpacked and applied to the target row
            - Any stored values are updated in the memory context.
        """
        target_row: TwoDARow | None = None

        if self.exclusive_column:
            if self.exclusive_column not in self.cells:
                msg = f"Exclusive column {self.exclusive_column} does not exists"
                raise WarningError(msg)

            exclusive_value = self.cells[self.exclusive_column].value(
                memory,
                twoda,
                None,
            )
            for row in twoda:
                if row.get_string(self.exclusive_column) != exclusive_value:
                    continue
                target_row = row

        if target_row is None:
            row_label: str = str(twoda.get_height()) if self.row_label is None else self.row_label
            index: int = twoda.add_row(row_label, {})
            self._row = target_row = twoda.get_row(index)
            target_row.update_values(self._unpack(self.cells, memory, twoda, target_row))
        else:
            cells: dict[str, str] = self._unpack(self.cells, memory, twoda, target_row)
            target_row.update_values(cells)

        for token_id, value in self.store_2da.items():
            memory.memory_2da[token_id] = value.value(memory, twoda, target_row)

        for token_id, value in self.store_tlk.items():
            memory.memory_str[token_id] = int(value.value(memory, twoda, target_row))


class CopyRow2DA(Modify2DA):
    """Copies the the row if the exclusive_column value doesn't already exist. If the row already exists, it simply modifies the existing line.

    Attributes:
    ----------
        identifier:
        target: Which row to copy.
        exclusive_column: Modify existing line if the same value already exists at this column.
        modifiers: For the row, sets a cell under column KEY to have the text VALUE.
    """

    def __init__(  # noqa: PLR0913
        self,
        identifier: str,
        target: Target,
        exclusive_column: str | None,
        row_label: str | None,
        cells: dict[str, RowValue],
        store_2da: dict[int, RowValue] | None = None,
        store_tlk: dict[int, RowValue] | None = None,
    ):
        self.identifier: str = identifier
        self.target: Target = target
        self.exclusive_column: str | None = exclusive_column or None
        self.row_label: str | None = row_label
        self.cells: dict[str, RowValue] = cells
        self.store_2da: dict[int, RowValue] = {} if store_2da is None else store_2da
        self.store_tlk: dict[int, RowValue] = {} if store_tlk is None else store_tlk

        self._row: TwoDARow | None = None

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(identifier={self.identifier!r}, "
            f"target={self.target!r}, exclusive_column={self.exclusive_column!r}, "
            f"row_label={self.row_label!r}, cells={self.cells!r}, "
            f"store_2da={self.store_2da!r}, store_tlk={self.store_tlk!r})"
        )

    def apply(self, twoda: TwoDA, memory: PatcherMemory):
        """Applies a CopyRow patch to a TwoDA.

        Args:
        ----
            twoda: TwoDA - The TwoDA to apply the patch to
            memory: PatcherMemory - The memory context

        Processing Logic:
        ----------------
            1. Searches for the source row in the TwoDA
            2. Determines if an existing target row should be used or a new row added
            3. Unpacks the cell values and updates/adds the target row
            4. Stores any 2DA or TLK values in the memory context.
        """
        source_row: TwoDARow | None = self.target.search(twoda, memory)
        target_row: TwoDARow | None = None
        row_label = str(twoda.get_height()) if self.row_label is None else self.row_label

        if source_row is None:
            msg = f"Source row cannot be None. row_label was '{row_label}'"
            raise WarningError(msg)

        if self.exclusive_column is not None:
            if self.exclusive_column not in self.cells:
                msg = f"Exclusive column {self.exclusive_column} does not exists"
                raise WarningError(msg)

            exclusive_value: str = self.cells[self.exclusive_column].value(
                memory,
                twoda,
                None,
            )
            for row in twoda:
                if row.get_string(self.exclusive_column) != exclusive_value:
                    continue
                target_row = row

        if target_row is not None:
            # If the row already exists (based on exclusive_column) then we copy all columns from source first
            # Copy all column values from source row to target row
            for header in twoda.get_headers():
                source_value = source_row.get_string(header)
                target_row.set_string(header, source_value)
            # Then update with the cells dictionary (which will override the copied values)
            cells = self._unpack(self.cells, memory, twoda, target_row)
            target_row.update_values(cells)
            self._row = target_row
        else:
            # Otherwise, we add the new row instead.
            index: int = twoda.copy_row(source_row, row_label, {})
            self._row = target_row = twoda.get_row(index)
            cells = self._unpack(self.cells, memory, twoda, target_row)
            target_row.update_values(cells)

        for token_id, value in self.store_2da.items():
            memory.memory_2da[token_id] = value.value(memory, twoda, target_row)

        for token_id, value in self.store_tlk.items():
            memory.memory_str[token_id] = int(value.value(memory, twoda, target_row))


class AddColumn2DA(Modify2DA):
    """Adds a column.

    The new cells are either given a default value or can be given a
    value based on what the row index or row label is.

    Attributes:
    ----------
        identifier:
        header: Label for the name column.
        default: Default value of cells if no specific value was specified.
        index_insert: For the new column, if the row index is KEY then set cell to VALUE.
        label_insert: For the new column, if the row label is KEY then set cell to VALUE.
    """

    def __init__(  # noqa: PLR0913
        self,
        identifier: str,
        header: str,
        default: str,
        index_insert: dict[int, RowValue],
        label_insert: dict[str, RowValue],
        store_2da: dict[int, str] | None = None,
    ):
        self.identifier: str = identifier
        self.header: str = header
        self.default: str = default
        self.index_insert: dict[int, RowValue] = index_insert
        self.label_insert: dict[str, RowValue] = label_insert
        self.store_2da: dict[int, str] = {} if store_2da is None else store_2da

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(identifier={self.identifier!r}, "
            f"header={self.header!r}, default={self.default!r}, "
            f"index_insert={self.index_insert}, label_insert={self.label_insert}, "
            f"store_2da={self.store_2da})"
        )

    def apply(self, twoda: TwoDA, memory: PatcherMemory):
        """Applies a AddColumn patch to a TwoDA.

        Args:
        ----
            twoda: TwoDA - The TwoDA to apply the patcher to
            memory: PatcherMemory - The memory object to store values

        Processing Logic:
        ----------------
            - Adds a column to the TwoDA with the patcher header
            - Sets the default value for all rows in the new column
            - Sets values in the new column based on index/label lookups
            - Stores values from the TwoDA in the memory based on token IDs.
        """
        twoda.add_column(self.header)
        for row in twoda:
            row.set_string(self.header, self.default)

        for row_index, row_value in self.index_insert.items():
            index_str: str = row_value.value(memory, twoda, None)
            this_row: TwoDARow | None = twoda.get_row(row_index)
            if this_row:
                this_row.set_string(self.header, index_str)
            else:
                msg = f"Could not find row {row_index} in {self.header}"
                raise WarningError(msg)

        for row_label, row_value in self.label_insert.items():
            label_str: str = row_value.value(memory, twoda, None)
            this_row = twoda.find_row(row_label)
            if this_row:
                this_row.set_string(self.header, label_str)
            else:
                msg = f"Could not find row {row_label} in {self.header}"
                raise WarningError(msg)

        for token_id, value in self.store_2da.items():
            if value.startswith("I"):
                cell: str = twoda.get_row(int(value[1:])).get_string(self.header)
                memory.memory_2da[token_id] = cell
            elif value.startswith("L"):
                row = twoda.find_row(value[1:])
                if row is None:
                    msg = f"Could not find row {value[1:]} in {self.header}"
                    raise WarningError(msg)
                cell = row.get_string(self.header)
                memory.memory_2da[token_id] = cell
            else:
                msg = f"store_2da dict has an invalid value at {token_id}: '{value}'"
                raise WarningError(msg)


# endregion


class Modifications2DA(PatcherModifications):
    hardcapped_row_limits = {
        "placeables.2da": 256,
        "upcrystals.2da": 256,
        "upgrade.2da": 32,
    }

    def __init__(self, filename: str):
        super().__init__(filename)
        self.modifiers: list[Modify2DA] = []

    def patch_resource(
        self,
        source: SOURCE_TYPES,
        memory: PatcherMemory,
        logger: PatchLogger,
        game: Game,
    ) -> bytes | Literal[True]:
        twoda: TwoDA = read_2da(source)
        self.apply(twoda, memory, logger, game)
        return bytes_2da(twoda)

    def apply(
        self,
        mutable_data: TwoDA,
        memory: PatcherMemory,
        logger: PatchLogger,
        game: Game,
    ):
        for row in self.modifiers:
            try:
                row.apply(mutable_data, memory)
            except Exception as e:  # noqa: PERF203, BLE001
                msg = f"{(e.__class__.__name__, str(e))} when patching the file '{self.saveas}'"
                RobustLogger().critical(str(e), exc_info=e)
                if isinstance(e, WarningError):
                    logger.add_warning(msg)
                    RobustLogger().debug(msg, exc_info=True)
                else:
                    logger.add_error(msg)
                    break
        if game.is_k2():  # k2 doesn't have hard limits?
            return

        # Exceeding row count maximums will break the game.
        twoda_row_limit = self.hardcapped_row_limits.get(self.saveas.lower())
        if twoda_row_limit is None:
            return
        cur_row_count = len(mutable_data._rows)
        rows_added = cur_row_count - twoda_row_limit
        if cur_row_count > twoda_row_limit:
            raise ValueError(
                f"{self.saveas} has a max row count of {twoda_row_limit}. Adding more will break the game. This mod attempted to add {rows_added} rows and have not been applied."
            )
